from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from itertools import chain, combinations_with_replacement

import numpy as np
from scipy.special import eval_legendre
from sklearn.base import TransformerMixin
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline


def _legendre(x, n):
    return eval_legendre(n, x)


def _var_multiplier(interaction_list):
    return np.prod([1. / (2 * el + 1) for el in interaction_list])


def _make_variance_terms(coefficients, interactions):
    return [_var_multiplier(interaction) * coefficient ** 2 for coefficient, interaction in
            zip(coefficients, interactions)]


class PolynomialBasis(TransformerMixin):
    """Transforms to Legendre Basis"""

    def __init__(self, num_degrees, basis):
        self.num_degrees = num_degrees
        self._basis = basis
        self.interactions = None

    def fit(self, x, y=None):
        dim = x.shape[1]

        # Enumerate all interactions between variables for the given number of degrees
        combinations = chain.from_iterable(combinations_with_replacement(range(dim),
                                                                         i + 1) for i in range(self.num_degrees))
        self.interactions = np.array([[k.count(d) for d in range(dim)] for k in combinations])

        return self

    def transform(self, x):
        # Apply basis function and multiply dimensions based on interaction level
        return np.prod(self._basis(x[..., np.newaxis], self.interactions.T[np.newaxis, ...]), axis=1)


class RangeScaler(TransformerMixin):
    """Scales to a new range from a given range"""

    def __init__(self, feature_range_from=None, feature_range_to=None):
        self.feature_range_from = feature_range_from
        self.feature_range_to = feature_range_to

    @staticmethod
    def _scale(val, from_lo, from_hi, to_lo, to_hi):
        return (((val - from_lo) * (to_hi - to_lo)) / (from_hi - from_lo)) + to_lo

    def fit(self, x, y=None):
        return self

    def transform(self, x):
        vals = self._scale(x,
                           self.feature_range_from[:, 0],
                           self.feature_range_from[:, 1],
                           self.feature_range_to[:, 0],
                           self.feature_range_to[:, 1])

        return vals


class PolynomialChaosExpansionModel(object):
    """Linear regression on orthogonal polynomial basis"""

    def __init__(self, num_degrees=1, ranges=None):

        if ranges is None:
            raise Exception()

        np_ranges_from = np.array(ranges)

        ranges_shape = np_ranges_from.shape

        if len(ranges_shape) != 2:
            raise Exception()

        if ranges_shape[1] != 2:
            raise Exception

        # Pipeline chains together transformation and regression
        self.pipeline = Pipeline([('Range_Scaling', RangeScaler(feature_range_from=np_ranges_from,
                                                                feature_range_to=np.array([[-1, 1]]))),
                                  ('Legendre_Basis', PolynomialBasis(num_degrees=num_degrees, basis=_legendre)),
                                  ('Linear_Model', LinearRegression(fit_intercept=True, copy_X=False))])

        self.num_degrees = num_degrees
        self.coefficients = None
        self.intercept = None
        self.interactions = None
        self.variance_terms = None

    def fit(self, x, y):
        self.pipeline.fit(x, y)

        self.coefficients = self.pipeline.named_steps['Linear_Model'].coef_
        self.intercept = self.pipeline.named_steps['Linear_Model'].intercept_
        self.interactions = self.pipeline.named_steps['Legendre_Basis'].interactions
        self.variance_terms = np.array([_make_variance_terms(coef_list,
                                                             self.interactions) for coef_list in self.coefficients])

        return self

    def predict(self, x):
        return self.pipeline.predict(x)

    def make_contributions(self, degree):
        """
        Compute Sobol indices for interactions up to specified degree.
        
        Returns indices in an order that matches PolynomialBasis.interactions,
        which uses combinations_with_replacement (like interaction_only=False).
        
        For the result to match _make_interactions output with interaction_only=True,
        the degree parameter must equal the model's num_degrees.
        
        Args:
            degree (int): Maximum interaction degree to include
            
        Returns:
            Array of Sobol indices, one per interaction term
        """
        # Identify which features participate in each interaction
        non_zero = self.interactions > 0
        
        # Select interactions with total degree <= specified degree
        range_selector = non_zero.sum(axis=1) <= degree
        
        # Group interactions by which features participate (regardless of polynomial degree)
        same_term = np.prod(
            ~(non_zero[..., np.newaxis] ^ non_zero.T[np.newaxis, ...]), 
            axis=1
        )
        
        # Sum variance contributions for grouped terms
        var_term_sum = np.dot(self.variance_terms, same_term)
        
        # Select only "first order" representation (each feature appears exactly once)
        first_selection = ~np.sum(
            (self.interactions == 1) ^ non_zero, 
            axis=1, 
            dtype='bool'
        )
        
        return var_term_sum[:, first_selection & range_selector]


if __name__ == "__main__":
    pass



import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as sts
from kosh.operators.core import KoshOperator
from ibis import mcmc, sensitivity


class KoshMCMC(KoshOperator):
    """
    Uses a kosh datasets as inputs and outputs and converts them into IBIS input
    and output variables. With other information the user provides such as a surrogate
    model, prior distribution, etc. MCMC sampling chains are
    """

    types = {"numpy": ["ibis/mcmc", ]}

    def __init__(self, *args, **options):
        """
        :param method: Type of MCMC sampling to use. Options are: "default_mcmc"
        and "discrepancy_mcmc".
        :type method: str
        :param observed: Kosh datasets of one or more arrays of the observed quantities of
        interest. The arrays should have measured quantity of interest as columns and observations
        as rows.
        :type observed: kosh datasets
        :param input_names: The unique names of the inputs.
        :type input_names: list
        :param low: The lower bound of the input variables.
        :type low: list
        :param high: The upper bound of the input variables.
        :type high: list
        :param proposal_sigmas: The standard deviation of the proposal distribution
        for each feature in the inputs.
        :type proposal_sigmas: list of floats
        :param surrogate_models:The surrogate models that represents the mapping of input
        values to output values. This model must be have a predict method that takes a
        numpy array and returns a numpy array (sklearn's fit/predict paradigm). The order of
        the models must match the order of the outputs.
        :type surrogate_models: list of trained models
        :param events: The event the output is associated with. There will be a unique
        event name for each row of the experimental or measured data. Length should match
        the number of rows in the outputs array.
        :type events: list of str
        :param quantities: The names of the quantity of interest.
        :type quantities: list of str
        :param observed_std: The error bound on the observed values.
        :type observed_std: list of float
        :param total_samples: The total number of sample points to return
        :type total_samples: int
        :param burn: The number of burn-in iterations
        :type burn: int
        :param every: The rate at which to save points. Saves every Nth iteration.
        :type every: int
        :param priors: The prior distributions of each input feature. The default is
        None which will result in using a uniform distribution over the whole range.
        :type priors: list of functions
        :param unscaled_low: The minimum of the unscaled input features.
        :type unscaled_low: list of floats
        :param unscaled_high: The maximum of the unscaled input features.
        :type unscaled_high: list of floats
        :param scaling: The type of scaling used on inputs. 'lin' or 'log'
        :type scaling: str
        :param start: The value at which to start the chains for each input.
        :type start: dict of str, float
        :param tau_start: The value at which to start the chains for "tau_{qoi}": the
        discrepancy variance for each quantity of interest. If used, length should be
        equal to the number of quantities of interest.
        :type tau_start: list of float
        :param n_chains: The number of chains to run in parallel
        :type n_chains: int
        :param prior_only: Whether to run the chain on just the prior distributions.
        :type prior_only: bool
        :param seed: The random seed for the chains
        :type seed: int
        :param scaled: Whether the inputs were scaled
        :type scaled: bool
        """

        super(KoshMCMC, self).__init__(*args, **options)
        self.options = options

    def operate(self, *inputs, format=None):

        # Read in input kosh datasets into one numpy array
        observed = inputs[0][:]
        for input_ in inputs[1:]:
            observed = np.append(X, input_[:], axis=0)
        try:
            Nsamp, Ndim = observed.shape
        except ValueError:
            observed = observed.reshape(-1, 1)
            Nsamp, Ndim = observed.shape

        method = self.options.get("method", "default_mcmc")
        input_names = self.options.get("input_names")
        input_dim = len(input_names)
        low = self.options.get("low")
        high = self.options.get("high")
        proposal_sigmas = self.options.get("proposal_sigmas")
        surrogate_models = self.options.get("surrogate_models")
        events = self.options.get("events")
        quantities = self.options.get("quantities")
        observed_std = self.options.get("observed_std")
        total_samples = self.options.get("total_samples")
        burn = self.options.get("burn")
        every = self.options.get("every")
        priors = self.options.get("priors", [sts.uniform.pdf]*input_dim)
        unscaled_low = self.options.get("unscaled_low", [None])
        unscaled_high = self.options.get("unscaled_high", [None])
        scaling = self.options.get("scaling", None)
        start = self.options.get("start", None)
        tau_start = self.options.get("tau_start", None)
        n_chains = self.options.get("n_chains", -1)
        prior_only = self.options.get("prior_only", False)
        seed = self.options.get("seed", None)

        # Verify Inputs
        if method == "default_mcmc":
            mcmc_obj = mcmc.DefaultMCMC()
        elif method == "discrepancy_mcmc":
            mcmc_obj = mcmc.DiscrepancyMCMC()
        else:
            msg = f"The MCMC method entered was '{method}'. Please choose "
            msg += "from 'default_mcmc', or 'discrepancy_mcmc'."
            raise ValueError(msg)

        # input_names, low, high, and proposal sigmas should have the same length
        error_msg = f"input_names, low, high, and proposal_sigmas have different lengths!"
        error_msg += f"\ninput_names: {len(input_names)}"
        error_msg += f"\nlow: {len(low)}"
        error_msg += f"\nhigh: {len(high)}"
        error_msg += f"\nproposal_sigmas: {len(proposal_sigmas)}"
        equal_sizes = len(input_names) == len(low) == len(high) == len(proposal_sigmas)
        assert equal_sizes, error_msg

        # Verify scaling and unscaled values are valid and correct length
        if scaling is not None and ((unscaled_low[0] is None) or (unscaled_high[0] is None)):
                msg = "If using scaling you must provide the unscaled low and high "
                msg += "values for each input parameter."
                raise ValueError(msg)
        if scaling is not None:
            error_msg = f"unscaled_low and unscaled high are a different length than input_names!"
            error_msg += f"\nunscaled_low: {len(unscaled_low)}"
            error_msg += f"\nunscaled_high: {len(unscaled_high)}"
            error_msg += f"\ninput_names: {len(input_names)}"
            equal_sizes = len(input_names) == len(unscaled_low) == len(unscaled_high)
            assert equal_sizes, error_msg

        # Verify output info
        error_msg = f"Number of quantities or interest (QOI) should match the "
        error_msg += f"number of observed_std."
        error_msg += f"\nQOI's: {Ndim}"
        error_msg += f"\nobserved_std: {len(observed_std)}"

        for i, event_i in enumerate(events):
            for j, qoi_j in enumerate(quantities):
                mcmc_obj.add_output(event=event_i,
                                    quantity=qoi_j,
                                    surrogate_model=surrogate_models[j],
                                    observed_value=observed[i,j],
                                    observed_std=observed_std[j],
                                    inputs=input_names)
        for i, name in enumerate(input_names):
            if (unscaled_low[0] is not None) and (unscaled_high[0] is not None):
                unscaled_low = unscaled_low[i]
                unscaled_high = unscaled_high[i]
            mcmc_obj.add_input(name=name,
                               low=low[i],
                               high=high[i],
                               proposal_sigma=proposal_sigmas[i],
                               prior=priors[i],
                               unscaled_low=unscaled_low,
                               unscaled_high=unscaled_high,
                               scaling=scaling)
        if tau_start is not None:
            for i, name in enumerate(quantities):
                start[f'tau_{name}'] = tau_start[i]
        mcmc_obj.run_chain(total=total_samples,
                           burn=burn,
                           every=every,
                           start=start,
                           n_chains=n_chains,
                           prior_only=prior_only,
                           seed=seed)

        return mcmc_obj


class KoshSensitivityPlots(KoshOperator):
    """
    Description
    """

    types = {"numpy": ["matplotlib/fig", ]}

    def __init__(self, *args, **options):
        """
        :param inputs: The input datasets. Kosh datasets of one or more arrays.
        The arrays should have features as columns and observation as rows.
        :type inputs: Kosh datasets
        :param input_names: The names of the inputs for plots
        :type inputs_names: list of str
        :param outputs: The output dataset. Rows correspond to rows in the feature data.
        The arrays should have features as columns and observation as rows.
        :type outputs: Kosh datasets
        :param output_names: The names of the output variables.
        :type output_names: list of str
        :param method: Plot type options are 'lasso', 'sensitivity', 'f_score',
        'mutual_info_score', 'pce_score', 'f_score_rank', 'mutual_info_rank',
        'pce_rank', 'f_score_network', or 'pce_network'.
        :type method: string
        :param degree: Maximum degree of interaction for an f_score plot
        :type degree: int
        :param model_degrees: Maximum degree of interaction for PCE model
        :type model_degrees: int
        :param surrogate_model: The surrogate model which has been fit to data
        :type surrogate_model: model with fit/predict functions
        :param input_ranges: Array-like of feature ranges. Each row is a length
        2 array of the lower and upper bounds.
        :type input_ranges: list of arrays
        :param num_plot_points: Number of points to plot on each dimension sweep
        :type num_plot_points: int
        :param num_seed_points: Number of points to use as default points
        :type num_seed_points: int
        :param seed: The random seed for the chains
        :type seed: int
        :param interaction_only: Whether to only include lowest powers of interaction or
        include higher powers for the f_score_plot.
        :type interaction_only: bool
        :param use_p_value: Whether to use p-values or raw F-score in the f_score_plot
        :type use_p_value: bool
        :param n_neighbors: How many neighboring bins to consider when estimating mutual information
        :type n_neighbors: int
        :param max_size: Maximum size of elements in plot. Measured in points
        :type max_size: float
        :param label_size: Font size of labels. Measured in points
        :type label_size: int
        :param alpha: Opacity of elements in plot.
        :type alpha: float
        :param save_plot: Whether to save the plot
        :type save_plot: bool
        """
        super(KoshSensitivityPlots, self).__init__(*args, **options)
        self.options = options

    def operate(self, *inputs, format=None):
        import string

        # Read in input kosh datasets into one numpy array
        X = inputs[0][:]
        for input_ in inputs[1:]:
            X = np.append(X, input_[:], axis=0)
        Ndim = X.shape[1]

        methods = ["oat_score", "oat_rank", "morris_score", "morris_rank", "sobol_score",
                   "sobol_rank", "lasso", "sensitivity", "f_score", "mutual_info_score",
                   "pce_score", "f_score_rank", "mutual_info_rank", "pce_rank", 
                   "f_score_network", "pce_network"]

        method = self.options.get("method")
        input_names = self.options.get("input_names",
                                       list(string.ascii_lowercase)[:Ndim])
        outputs = self.options.get("outputs")
        output_names = self.options.get("output_names", "response")
        show_both = self.options.get("show_both", True)
        rank_by = self.options.get("rank_by", "mu_star")
        statistic = self.options.get("statistic", "max")
        index_type = self.options.get("index_type", "first_order")
        include_second_order = self.options.get("include_second_order", False)
        surrogate_model = self.options.get("surrogate_model")
        input_ranges = self.options.get("input_ranges")
        num_plot_points = self.options.get("num_plot_points", 100)
        num_seed_points = self.options.get("num_seed_points", 5)
        seed = self.options.get("seed", 2024)
        degree = self.options.get("degree", 1)
        model_degrees = self.options.get("model_degrees", 1)
        interaction_only = self.options.get("interaction_only", True)
        use_p_value = self.options.get("use_p_value", False)
        n_neighbors = self.options.get("n_neighbors", 3)
        max_size = self.options.get("max_size", 10.0)
        label_size = self.options.get("label_size", 10)
        alpha = self.options.get("alpha", 0.5)
        save_plot = self.options.get("save_plot", True)

        # Read in output dataset
        Y = np.array(outputs[:])

        if method == "oat_score":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.oat_score_plot(ax, X, Y, input_names, output_names,
                                       statistic=statistic)

        elif method == "oat_rank":
            fig, ax = plt.subplots(1, 1)  # Single subplot only
            sensitivity.oat_rank_plot(ax, X, Y, input_names, output_names,
                                       statistic=statistic)

        elif method == "morris_score":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.morris_score_plot(ax, X, Y, input_names, output_names,
                                       show_both=show_both, degree=degree,
                                       interaction_only=interaction_only)

        elif method == "morris_rank":
            fig, ax = plt.subplots(1, 1)  # Single subplot only
            sensitivity.morris_rank_plot(ax, X, Y, input_names, output_names,
                                       rank_by=rank_by, show_both=show_both, 
                                       degree=degree, interaction_only=interaction_only)

        elif method == "sobol_score":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.sobol_score_plot(ax, X, Y, input_names, output_names,
                                       index_type=index_type,
                                       include_second_order=include_second_order)

        elif method == "sobol_rank":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.sobol_rank_plot(ax, X, Y, input_names, output_names,
                                       index_type=index_type,
                                       include_second_order=include_second_order)

        elif method == "lasso":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.lasso_path_plot(ax, X, Y, input_names, output_names,
                                        degree=degree, method='lasso')
        elif method == "sensitivity":
            fig, ax = plt.subplots(len(input_names), len(output_names))
            sensitivity.sensitivity_plot(ax, surrogate_model, input_names,
                                         output_names, input_ranges,
                                         num_plot_points=num_plot_points,
                                         num_seed_points=num_seed_points, seed=seed)
        elif method == "f_score":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.f_score_plot(ax, X, Y, input_names, output_names,
                                     degree=degree, interaction_only=interaction_only,
                                     use_p_value=use_p_value)
        elif method == "mutual_info_score":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.mutual_info_score_plot(ax, X, Y, input_names, output_names,
                                               n_neighbors=n_neighbors)
        elif method == "pce_score":
            fig, ax = plt.subplots(1, len(output_names))
            sensitivity.pce_score_plot(ax, X, Y, input_names, output_names,
                                       input_ranges, degree=degree,
                                       model_degrees=model_degrees)
        elif method == "f_score_rank":
            fig, ax = plt.subplots(1, 1)
            sensitivity.f_score_rank_plot(ax, X, Y, input_names, output_names,
                                          degree=degree, interaction_only=interaction_only,
                                          use_p_value=use_p_value)
        elif method == "mutual_info_rank":
            fig, ax = plt.subplots(1, 1)
            sensitivity.mutual_info_rank_plot(ax, X, Y, input_names, output_names,
                                              n_neighbors=n_neighbors)
        elif method == "pce_rank":
            fig, ax = plt.subplots(1, 1)
            sensitivity.pce_rank_plot(ax, X, Y, input_names, output_names, input_ranges,
                                      degree=degree, model_degrees=model_degrees)
        elif method == "f_score_network":
            fig, ax = plt.subplots(len(output_names), (degree-1))
            sensitivity.f_score_network_plot(ax, X, Y, input_names, output_names,
                                             degree=degree, max_size=max_size,
                                             label_size=label_size, alpha=alpha)
        elif method == "pce_network":
            fig, ax = plt.subplots(len(output_names), (degree-1))
            sensitivity.pce_network_plot(ax, X, Y, input_names, output_names,
                                         input_ranges, degree=degree,
                                         model_degrees=model_degrees, max_size=max_size,
                                         label_size=label_size, alpha=alpha)
        else:
            msg = f"Method should be one of the following: {methods}."
            msg += f"Was given {method}."
            raise ValueError(msg)

        if save_plot:
            names = "_".join(output_names)
            fileName = f"{names}_{method}_plot.png"
            fig.savefig(fileName)

        return fig



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np


class Likelihood(object):

    @staticmethod
    def combine_weights(weights, **kwargs):
        raise NotImplementedError


class Intersection(Likelihood):
    """
    Intersection
    """

    @staticmethod
    def combine_weights(weights, **kwargs):
        """
        :param weights:
        :param kwargs:
        :return:
        """
        if len(weights) == 0:
            return np.ones(1)

        np_combined_weights = np.prod(weights, axis=0)
        # raises ValueError if weight arrays are not the same length

        return np.divide(np_combined_weights, np.sum(np_combined_weights))


class StatFiltering(Likelihood):
    """
    Statistical Filtering
    """

    @staticmethod
    def combine_weights(weights, **kwargs):
        """
        :param weights:
        :param kwargs:
        :return:
        """
        if len(weights) == 0:
            return np.ones(1)

        np_combined_weights = np.sum(weights, axis=0)
        # raises ValueError if weight arrays are not the same length

        f_max_weight = np_combined_weights.max()

        if 'chi' in kwargs:
            chi = kwargs['chi']
        else:
            # compute chi
            np_max_weights = np.max(weights, axis=1)
            chi = np.log(1.0 - np.mean(np_max_weights) / f_max_weight)
            chi = max(-1.0 / chi, 1.0)
            # print "chi: {}".format(chi)

        return (np_combined_weights / f_max_weight) ** chi



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np
import scipy.ndimage
from mpl_toolkits.axes_grid1 import make_axes_locatable


def likelihood_plot(ax, prior_points, post_weights=None, post_points=None, exp_value=None,
                    conf_level=None, bins=10, density=True):
    """
        Shows the prior and posterior likelihood of a set of points.

        :parameter ax: The Axes to plot to
        :type ax: matplotlib axes object

        :parameter prior_points: A set of points
        :type prior_points: numpy array

        :parameter post_weights: The posterior weights of the points
        :type post_weights: numpy array

        :parameter post_points: The posterior set of points
        :type post_points: numpy array

        :parameter exp_value: Experimental value
        :type exp_value: float

        :parameter conf_level: Confidence levels
        :type conf_level: float

        :parameter bins: Number of bins to partition into
        :type bins: int

        :parameter density: Whether to create a plot that integrates to 1
        :type density: bool
    """
    # if not len(prior) == len(posterior):
    #    raise Exception()

    if post_weights is not None and post_points is not None:
        raise Exception('Cannot use both weights and points at the same time')

    if post_weights is not None and np.any(post_weights < 0):
        raise Exception('Posterior weights cannot be negative')

    # Partition points into bins
    binned_prior, prior_x = np.histogram(prior_points, bins=bins, density=density)
    if post_weights is not None:
        binned_posterior, posterior_x = np.histogram(prior_points, bins=bins, weights=post_weights, density=density)
    elif post_points is not None:
        binned_posterior, posterior_x = np.histogram(post_points, bins=bins, density=density)

    prior_x = 0.5 * (prior_x[0:-1] + prior_x[1:])
    if post_weights is not None or post_points is not None:
        posterior_x = 0.5 * (posterior_x[0:-1] + posterior_x[1:])

    # Set y-axis limits
    max_prior = np.max(binned_prior)
    if post_weights is not None or post_points is not None:
        max_posterior = np.max(binned_posterior)
        upper_limit_y = max(max_prior, max_posterior) * 1.2
    else:
        upper_limit_y = max_prior * 1.2
    ax.set_ylim([0, upper_limit_y])

    # Set x-axis limits
    range_x = [np.min(prior_x), np.max(prior_x)]
    if exp_value is not None:
        lower_limit_x = min(exp_value - (.1 * (range_x[1] - exp_value)), range_x[0])
        upper_limit_x = max(exp_value + (.1 * (exp_value - range_x[0])), range_x[1])
    else:
        lower_limit_x = range_x[0]
        upper_limit_x = range_x[1]
    ax.set_xlim([lower_limit_x, upper_limit_x])

    # Plot distributions
    ax.fill_between(prior_x, binned_prior, color='blue', alpha=.25, label='Prior')
    if post_weights is not None or post_points is not None:
        ax.fill_between(posterior_x, binned_posterior, color='green', alpha=.25, label='Posterior')

    # Plot experimental value
    if exp_value is not None:
        ax.vlines(float(exp_value), 0, upper_limit_y, color='red', linestyle='dashed', label='Exp. Value')

    if conf_level is not None:
        if conf_level < 0 or conf_level > 100:
            raise Exception("Confidence level must be between 0 and 100")

        ls_alpha = [(100 - conf_level) / 2., conf_level / 2. + 50]

        # Plot confidence intervals
        ls_prior_index = _get_conf_level_index(binned_prior, ls_alpha)
        ax.plot(prior_x[ls_prior_index], binned_prior[ls_prior_index], color='blue', marker='o', linestyle='',
                label="Prior predictive {}% CI".format(conf_level))
        if post_weights is not None or post_points is not None:
            ls_posterior_index = _get_conf_level_index(binned_posterior, ls_alpha)
            ax.plot(posterior_x[ls_posterior_index], binned_posterior[ls_posterior_index], color='green', marker='o',
                    linestyle='',
                    label="Posterior predictive {}% CI".format(conf_level))
    ax.set_yticklabels([])
    ax.set_ylabel('density')

    if post_points is None and post_weights is None:
        return prior_x, binned_prior
    else:
        return prior_x, binned_prior, posterior_x, binned_posterior

def _get_conf_level_index(y, conf_levels):
    """
        Gives a the particular indices that given confidence levels occur at

        :parameter y: Set of points
        :type y: numpy array

        :parameter conf_levels: Confidence levels
        :type conf_levels: list of int

        :return: Indices of confidence levels
        :rtype: list of int
    """
    cdf = np.cumsum(y)
    ls_index = [np.searchsorted(cdf, 0.01 * l * cdf[-1], side='right') for l in conf_levels]

    # If the desired value ends up so that it should be inserted after the
    # last element in the list then we need to adjust the indices to prevent
    # index errors.
    num_cdf = len(cdf)
    ls_index = [i if i < num_cdf else (num_cdf - 1) for i in ls_index]
    return ls_index


def slice_plot_2d(ax, model, xvar, yvar, default, num_points=10, fig=None):
    """
        Shows a 2-D slice of a surrogate model

        :parameter ax: The Axes to plot to
        :type ax: matplotlib axes object

        :parameter model: Surrogate model to evaluate
        :type model: SurrogateModel

        :parameter xvar: The name of the variable to plot on x-axis
        :type xvar: str

        :parameter yvar: The name of the variable to plot on y-axis
        :type yvar: str

        :parameter default: the default point at which to make the slice
        :type default: numpy array

        :parameter num_points: The number of points in sliced axes to plot
        :type num_points: int

        Returns:  cax, X, Y, Z

    """
    x_names = model.X_names
    x_ranges = model.X_range

    xvar_index = np.where(x_names == xvar)[0][0]
    yvar_index = np.where(x_names == yvar)[0][0]

    xvar_range = x_ranges[xvar_index]
    yvar_range = x_ranges[yvar_index]

    xvar_points = np.linspace(xvar_range[0], xvar_range[1], num_points)
    yvar_points = np.linspace(yvar_range[0], yvar_range[1], num_points)

    X, Y = np.meshgrid(xvar_points, yvar_points)

    X_flat = X.flatten()
    Y_flat = Y.flatten()

    points = np.repeat([default], num_points ** 2, 0).T
    points[xvar_index] = X_flat
    points[yvar_index] = Y_flat

    Z = model.predict(points.T)[0].reshape(num_points, num_points)

    cax = ax.contourf(X, Y, Z)
    ax.set_title("%s slice plot" % model.Y_names[0])
    ax.set_xlabel(xvar)
    ax.set_ylabel(yvar)

    if fig is not None:
        fig.colorbar(cax)
    return cax, X, Y, Z

def slice_plot_2d_std(ax, model, xvar, yvar, default, num_points=10, fig=None):
    """
        Shows a 2-D slice of a surrogate model's predicted standard deviation

        :parameter ax: The Axes to plot to
        :type ax: matplotlib axes object

        :parameter model: Surrogate model to evaluate
        :type model: SurrogateModel

        :parameter xvar: The name of the variable to plot on x-axis
        :type xvar: str

        :parameter yvar: The name of the variable to plot on y-axis
        :type yvar: str

        :parameter default: the default point at which to make the slice
        :type default: numpy array

        :parameter num_points: The number of points in sliced axes to plot
        :type num_points: int

        Returns: cax, X, Y, Z

    """
    x_names = model.X_names
    x_ranges = model.X_range

    xvar_index = np.where(x_names == xvar)[0][0]
    yvar_index = np.where(x_names == yvar)[0][0]

    xvar_range = x_ranges[xvar_index]
    yvar_range = x_ranges[yvar_index]

    xvar_points = np.linspace(xvar_range[0], xvar_range[1], num_points)
    yvar_points = np.linspace(yvar_range[0], yvar_range[1], num_points)

    X, Y = np.meshgrid(xvar_points, yvar_points)

    X_flat = X.flatten()
    Y_flat = Y.flatten()

    points = np.repeat([default], num_points ** 2, 0).T
    points[xvar_index] = X_flat
    points[yvar_index] = Y_flat

    Z = model.predict(points.T, return_std=True)[1]

    if Z is None:
        raise ValueError("Surrogate Model did not return standard deviation")

    Z = Z.reshape(num_points, num_points)

    cax = ax.contourf(X, Y, Z)
    ax.set_title("%s slice plot [std]" % model.Y_names[0])
    ax.set_xlabel(xvar)
    ax.set_ylabel(yvar)

    if fig is not None:
        fig.colorbar(cax)

    return cax, X, Y, Z

def contour_plot(ax, x, y, bins=10, weights=None, fig=None):
    """
        Shows the shaded contours of a 2-D set of points, optionally weighted with posterior probabilities

        :parameter ax: The Axes to plot to
        :type ax: matplotlib axes object

        :parameter x: A set of points on the x-axis
        :type x: numpy array

        :parameter y: A set of points on the y-axis
        :type y: numpy array

        :parameter bins: Number of bins to partition into along each axis. Total number of bins will be N*N
        :type bins: int

        :parameter weights: The posterior weights of the points
        :type weights: numpy array
    
        Returns: cax, X, Y, Z

    """
    c_levels = [0.025, 0.25, 0.50, 0.75, 0.975]
    c_colors = ['blue', 'green', 'black', 'green', 'blue']
    c_styles = ['solid', 'dashed', 'solid', 'dashed', 'solid']

    Z, X, Y = np.histogram2d(x, y, bins=bins, weights=weights, density=True)
    X = 0.5 * (X[0:-1] + X[1:])
    Y = 0.5 * (Y[0:-1] + Y[1:])
    cax = ax.contourf(X, Y, Z.T)

    if fig is not None:
        fig.colorbar(cax)

    return cax, X, Y, Z

    # ax.contour(X, Y, Z,
    #           levels=c_levels,
    #           colors=c_colors,
    #           linestyles=c_styles)


def scatter_plot(ax, x, y, num_points=None, weights=None, seed=None):
    """
        Shows the scatter plot of a 2-D set of points, optionally culled by posterior probabilities

        :parameter ax: The Axes to plot to
        :type ax: matplotlib axes object

        :parameter x: A set of points on the x-axis
        :type x: numpy array

        :parameter y: A set of points on the y-axis
        :type y: numpy array

        :parameter num_points: Number of points to plot
        :type num_points: int

        :parameter weights: The posterior weights of the points
        :type weights: numpy array
    """
    rng = np.random.default_rng(seed=seed)

    if num_points is None:
        ax.plot(x, y, 'o')
        return x, y
    else:
        if weights is not None:
            idx = rng.choice(range(len(x)), num_points, p=weights/weights.sum())
        else:
            idx = rng.choice(range(len(x)), num_points)
        ax.plot(x[idx], y[idx], '.')
        return x[idx], y[idx]


def box_plot(ax, prior_preds, exp_obs, exp_std, posterior_wts=None, posterior_preds=None, add_uncert=None, num_pts=100,
             num_bins=100, seed=None):
    """
        Plots 3 box plots

        Shows 3 box plots representing the prior distribution, posterior distribution,
        and posterior distribution + additional uncertainties. Also shows a histogram of
        posterior distribution + additional uncertainties.

        :parameter ax: The Axes to plot to
        :type ax: matplotlib axes object

        :parameter prior_preds: A set of points
        :type prior_preds: numpy array

        :parameter posterior_wts: The posterior weights of the points
        :type posterior_wts: numpy array

        :parameter exp_obs: Experimental observed value
        :type exp_obs: float

        :parameter exp_std: Experimental standard deviation
        :type exp_std: float

        :parameter add_uncert: Additional uncertainties
        :type add_uncert: list of 2-tuple of float

        :parameter num_pts: Number of points to sample from probability models
        :type num_pts: int

        :parameter num_bins: Number of histogram bins
        :type num_bins: int

        :parameter seed: Random seed
        :type seed: int
    """

    def model_pred_dists(points, weights, std=None, bins=10, std_mltpl=10, add_uncert=None):
        """
            Convolution step

            :parameter points: A set of points
            :type points: numpy array

            :parameter weights: A set of posterior weights
            :type weights: numpy array

            :parameter std: Experimental standard deviation
            :type std: float

            :parameter bins: Number of histogram bins
            :type bins: int

            :parameter std_mltpl: Standard deviation multiplier
            :type std_mltpl: float

            :parameter add_uncert: Additional uncertainties
            :type add_uncert: list of 2-tuple of float

            :return: A set of edges and heights. Optionally a set of heights associated with additional uncertainties
            :rtype: numpy array, numpy array, list of numpy array

            Returns: mean and std for model_prior, model_post and pred_post

        """
        if add_uncert is None:
            add_uncert = []

        mu = np.mean(points)
        low = mu - std_mltpl * std
        upp = mu + std_mltpl * std
        gauss_filter = scipy.ndimage.gaussian_filter1d

        heights, edges = np.histogram(points, bins=bins, density=True, range=[low, upp], weights=weights)
        dy = (edges[-1] - edges[0]) / bins

        ls_heights = [heights]
        for _, f_std in add_uncert:
            ls_heights.append(gauss_filter(ls_heights[-1], (mu * f_std / 100.0) / dy, order=0, mode='constant'))

        return edges, heights, ls_heights

    if posterior_wts is not None and posterior_preds is not None:
        raise Exception('Cannot use both weights and points at the same time')
    elif posterior_wts is None and posterior_preds is None:
        raise Exception('Must specify one of posterior_wts or posterior_preds')

    if add_uncert is None:
        add_uncert = []

    prior_edges, mod_prior_heights, pred_prior_heights = model_pred_dists(prior_preds,
                                                                          np.ones(prior_preds.shape),
                                                                          std=exp_std,
                                                                          bins=num_bins)
    dy = prior_edges[1] - prior_edges[0]

    if seed is not None:
        np.random.seed(seed + 1)
    np_model_prior = np.random.choice(prior_edges[0:-1] + 0.5 * dy, num_pts,
                                      p=mod_prior_heights / mod_prior_heights.sum())
    model_prior_mean = np.mean(np_model_prior)
    model_prior_std = np.std(np_model_prior)

    if posterior_wts is not None:
        post_edges, mod_post_heights, ls_pred_post_heights = model_pred_dists(prior_preds,
                                                                              posterior_wts,
                                                                              std=exp_std,
                                                                              bins=num_bins,
                                                                              add_uncert=add_uncert)
    elif posterior_preds is not None:
        post_edges, mod_post_heights, ls_pred_post_heights = model_pred_dists(posterior_preds,
                                                                              np.ones(posterior_preds.shape),
                                                                              std=exp_std,
                                                                              bins=num_bins,
                                                                              add_uncert=add_uncert)

    pred_post_heights = ls_pred_post_heights[-1]

    dy = post_edges[1] - post_edges[0]

    if seed is not None:
        np.random.seed(seed + 2)
    np_model_post = np.random.choice(post_edges[0:-1] + 0.5 * dy, num_pts, p=mod_post_heights / mod_post_heights.sum())
    model_post_mean = np.mean(np_model_post)
    model_post_std = np.std(np_model_post)

    np_pred_post = np.random.choice(post_edges[0:-1] + 0.5 * dy, num_pts,
                                    p=pred_post_heights / pred_post_heights.sum())
    pred_post_mean = np.mean(np_pred_post)
    pred_post_std = np.std(np_pred_post)

    ls_dists = [np_model_prior, np_model_post, np_pred_post]

    ls_x_lbl = ['prior\npredictive',
                'posterior\npredictive',
                'posterior\npredictive\n+ addl_uncert\n({})'.format(
                    '\n+ '.join(['{}'.format(tpUnc[0]) for tpUnc in add_uncert]))]
    pos = list(range(1, len(ls_dists) + 1))

    if model_prior_mean > 1.0:
        ax.annotate("$\\mu$={:.2f}\n$\\sigma$={:.2f}".format(model_prior_mean, model_prior_std),
                    (.05, .8), xycoords='axes fraction')
        ax.annotate("$\\mu$={:.2f}\n$\\sigma$={:.2f}".format(model_post_mean, model_post_std),
                    (.35, .8), xycoords='axes fraction')
        ax.annotate("$\\mu$={:.2f}\n$\\sigma$={:.2f}".format(pred_post_mean, pred_post_std),
                    (.65, .8), xycoords='axes fraction')
    else:
        ax.annotate("$\\mu$={:.4f}\n$\\sigma$={:.4f}".format(model_prior_mean, model_prior_std),
                    (.05, .8), xycoords='axes fraction')
        ax.annotate("$\\mu$={:.4f}\n$\\sigma$={:.4f}".format(model_post_mean, model_post_std),
                    (.35, .8), xycoords='axes fraction')
        ax.annotate("$\\mu$={:.4f}\n$\\sigma$={:.4f}".format(pred_post_mean, pred_post_std),
                    (.65, .8), xycoords='axes fraction')

    ax.boxplot(ls_dists, sym='', whis=1.5, patch_artist=True, positions=pos, medianprops={'color': 'k'})
    ax.set_xticklabels(ls_x_lbl, rotation=30)

    divider = make_axes_locatable(ax)
    ax_histogram = divider.append_axes('right', size='10%', pad='2%', sharey=ax)
    yy = np.ravel(list(zip(post_edges, post_edges)))
    xx = np.ravel(list(zip([0.] + list(pred_post_heights), list(pred_post_heights) + [0.])))
    ax_histogram.fill_betweenx(yy, xx)

    return (model_prior_mean, model_prior_std,
            model_post_mean, model_post_std,
            pred_post_mean, pred_post_std)



# Copyright (c) 2004 - 2014, Lawrence Livermore National Security, LLC.
# All rights reserved.
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np
from sklearn.mixture import GaussianMixture


class Filter(object):
    def __init__(self):
        object.__init__(self)

    @staticmethod
    def get_type():
        raise NotImplementedError

    @staticmethod
    def get_weights(output_values, observed_value, observed_std, **kwargs):
        raise NotImplementedError


class GaussianFilter(Filter):
    """
    Gaussian Likelihood
    """

    # def __init__(self):
    #    Filter.__init__(self)

    @staticmethod
    def get_type():
        return 'gaussian'

    @staticmethod
    def get_weights(output_values, observed_value, observed_std, **kwargs):
        """
        :param output_values:
        :param observed_value:
        :param observed_std:
        :param kwargs:
        :return:
        """
        if observed_std < 0:
            raise ValueError("Standard deviation cannot be less than zero."
                             "(experiment_stddev: {})".format(observed_std))

        if 'sigma_cut' in kwargs:
            sigma_cut = kwargs['sigma_cut']
            if sigma_cut < 0:
                raise ValueError("Sigma cutoff cannot be less than zero."
                                 "(sigma_cut: {})".format(sigma_cut))
        else:
            sigma_cut = 1.0

        np_weights = np.exp(-0.5 * (((output_values - observed_value) / observed_std) ** 2.0))

        return np.where(np.logical_and(output_values >= (observed_value - (sigma_cut * observed_std)),
                                       output_values <= (observed_value + (sigma_cut * observed_std))),
                        np_weights, 0.0)


class LogGaussianFilter(Filter):

    @staticmethod
    def get_type():
        return 'log-gaussian'

    @staticmethod
    def get_weights(output_values, observed_value, observed_std, **kwargs):
        """
        Log gaussian normal pdf

        optimized for speed

        :param output_values:
        :param observed_value:
        :param observed_std:
        :param kwargs:
        :return:
        """

        return -0.5 * (((output_values - observed_value) / observed_std) ** 2.0)


class StudentTFilter(Filter):
    """
    Student T Filter
    Based on Student's T distribution. Has fatter tails than the Gaussian Filter.
    Approaches Gaussian Filter as df -> inf
    """

    @staticmethod
    def get_type():
        return 'student-t'

    @staticmethod
    def get_weights(output_values, observed_value, observed_std, **kwargs):
        """
        :param output_values:
        :param observed_value:
        :param observed_std:
        :param kwargs:
        :return:
        """
        if observed_std < 0:
            raise ValueError("Standard deviation cannot be less than zero."
                             "(experiment_stddev: {})".format(observed_std))

        if 'df' in kwargs:
            df = float(kwargs['df'])
            if df <= 0:
                raise ValueError("Degrees of freedom must be positive."
                                 "(df: {})".format(df))
        else:
            df = 1.0

        if 'sigma_cut' in kwargs:
            sigma_cut = kwargs['sigma_cut']
            if sigma_cut < 0:
                raise ValueError("Sigma cutoff cannot be less than zero."
                                 "(sigma_cut: {})".format(sigma_cut))
        else:
            sigma_cut = 1.0

        np_weights = np.power(1.0 + (((output_values - observed_value) / observed_std) ** 2.0) / df, -(df + 1.0) / 2.0)

        return np.where(np.logical_and(output_values >= (observed_value - (sigma_cut * observed_std)),
                                       output_values <= (observed_value + (sigma_cut * observed_std))),
                        np_weights, 0.0)


class TophatFilter(Filter):

    @staticmethod
    def get_type():
        return 'top-hat'

    @staticmethod
    def get_weights(output_values, observed_value, observed_std, **kwargs):
        """
        :param output_values:
        :param observed_value:
        :param observed_std:
        :param kwargs:
        :return:
        """

        if 'sigma_cut' in kwargs:
            sigma_cut = kwargs['sigma_cut']
            if sigma_cut < 0:
                raise ValueError("Sigma cutoff cannot be less than zero."
                                 "(sigma_cut: {})".format(sigma_cut))
        else:
            sigma_cut = 1.0

        if observed_std < 0:
            raise ValueError("Standard deviation cannot be less than zero."
                             "(experiment_stddev: {})".format(observed_std))

        return np.where(np.logical_and(output_values >= (observed_value - sigma_cut),
                                       output_values <= (observed_value + sigma_cut)),
                        1.0, 0.0)


class LogTophatFilter(Filter):

    @staticmethod
    def get_type():
        return 'log-top-hat'

    @staticmethod
    def get_weights(output_values, observed_value, observed_std, **kwargs):
        return np.where(np.logical_and(output_values >= (observed_value - observed_std),
                                       output_values <= (observed_value + observed_std)),
                        0.0, float('-inf'))


class GaussianMixtureFilter(Filter):

    @staticmethod
    def get_type():
        return 'gaussian-mixture'

    @staticmethod
    def get_weights(output_values, observed_value, observed_std, **kwargs):
        """
        :type array-like [[]]
        :param output_values:

        :type array-like [[]]
        :param observed_value:

        :type array-like [[]]
        :param observed_std:

        :param kwargs:

        :type numpy array
        :return:

        The Gaussian Mixture Filter both filters and combines weights all at once.
        Inputs are expected to be multidimensional and in list (i.e. 2d arrays)
        """

        np_x_data = None

        for obs, std in zip(observed_value, observed_std):

            np_cov = np.diag(std) ** 2

            np.random.seed(20150515)
            dx = np.random.multivariate_normal(obs, np_cov, 30).T
            if np_x_data is None:
                np_x_data = np.c_[dx]
            else:
                np_x_data = np.c_[np_x_data, dx]

        # Find optimal number of components by minimizing Bayesian Information Criteria
        ls_bic_scores = []
        for i in range(len(observed_value)):
            gmm = GaussianMixture(n_components=i + 1, covariance_type='full')
            gmm.fit(np_x_data.T)
            ls_bic_scores += [gmm.bic(np_x_data.T)]
        num_min_bic = ls_bic_scores.index(min(ls_bic_scores))

        # Fit model
        gmm = GaussianMixture(n_components=num_min_bic + 1, covariance_type='full', n_init=10)
        gmm.fit(np_x_data.T)

        # Weight is the probability of point being in the model
        np_log_prob = np.array([gmm.score(val[np.newaxis, :]) for val in output_values])

        return np.exp(np_log_prob - np_log_prob.max())



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

if __name__ == "__main__":
    import matplotlib

    matplotlib.use('TKagg')

import functools
import logging
import multiprocessing
import sys
import time
import traceback
from copy import deepcopy

import matplotlib.pyplot as plt
import mpl_toolkits.axes_grid1
import numpy as np
import scipy.stats as sts

from ibis import mcmc_diagnostics
import warnings

def time_it(func):
    def timed(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()

        s = "    {}: {:.6f} s".format(func.__name__, (end - start))
        print(s)
        logging.info(s)
        return result

    return timed


class MCMC(object):
    """
        An interface for setting up and running Markov Chain Monte Carlo.

        An interface for setting up and running Markov Chain Monte Carlo on a set of unobserved inputs and observed
        outputs with surrogate models mapping inputs to outputs.
        Sub classes must implement _run_chain()
    """

    def __init__(self):
        self.outputs = {}
        self.inputs = {}
        self.chains = None
        self.acceptance_rate = None
        self.log_prob = None

        logging.info('Instantiated {}'.format(type(self).__name__))

    def add_output(self, event, quantity, surrogate_model, observed_value, observed_std, inputs):
        """
            Add an output variable to the set of variables

            An output variable represents an observed output that depends on some unobserved inputs. The function that
            maps inputs to outputs is approximated with a surrogate model.

            :parameter name: The unique name of the output
            :type name: str

            :parameter event: The event the output is associated with
            :type event: str

            :parameter quantity: The physical quantity of the output
            :type quantity: str

            :parameter surrogate_model: The surrogate model that represents the mapping of input values to output
                                        values. This model must be have a predict method that takes a numpy array and
                                        returns a numpy array (sklearn's fit/predict paradigm)

            :parameter observed_value: The observed experimental value
            :type observed_value: float

            :parameter observed_std: The error bound on the observed value
            :type observed_std: float

            :parameter inputs: The ordered names of the input variables that map to the output
            :type inputs: list of str
        """
        self.outputs[event + '_' + quantity] = OutputVariable(event, quantity, surrogate_model, observed_value,
                                                              observed_std, inputs)

        logging.info('\nAdded Output\nEvent: {}\nQuantity: {}\nObs Mean: {}\nObs Std: {}\n'
                     'Inputs: {}\nSurrogate Model: {}'.format(event, quantity,
                                                              observed_value,
                                                              observed_std, inputs,
                                                              surrogate_model))

    def remove_output(self, name):
        """
            Remove an output variable from the set of variables

            :parameter name: The name of the output variable to remove
            :type name: str
        """

        del self.outputs[name]

    def add_input(self, name, low, high, proposal_sigma, prior=None, unscaled_low=None, unscaled_high=None, scaling=None):
        """
            Add an input variable to the set of variables

            An input variable represents an unobserved input for some unobserved outputs. An input can be used for 1 or
            many outputs

            :parameter name: The unique name of the input
            :type name: str

            :parameter low: The lower bound on the value this variable can take
            :type low: float

            :parameter high: The upper bound on the value this variable can take
            :type high: float

            :parameter proposal_sigma: The standard deviation of the proposal distribution for this variable
            :type proposal_sigma: float

            :parameter prior: The prior distribution of the variable.
                              If None, a uniform distribution over the whole range is used
            :type prior: function
        """
        self.inputs[name] = InputVariable(name, low, high, proposal_sigma, prior, unscaled_low, unscaled_high, scaling)
        logging.info('\nAdded Input\nName: {}\nLower Bound: {}\nUpper Bound: {}\n'
                     'Proposal Sigma: {}\nPrior Distribution: {}\n'
                     'Unscaled Lower Bound: {}\nUnscaled Upper Bound: {}\n'
                     'Scaling: {}'.format(name, low, high, proposal_sigma, prior, unscaled_low, unscaled_high, scaling))

    def remove_input(self, name):
        """
            Remove an input variable from the set of variables

            :parameter name: The name of the input variable to remove
            :type name: str
        """

        del self.inputs[name]

    @time_it
    def run_chain(self, total, burn, every, start=None, n_chains=-1, prior_only=False, seed=None):
        """
            Runs the MCMC Algorithm

            Number of Iterations = n_chains*(total*every + burn)

            :parameter total: The total number of sample points to return
            :type total: int

            :parameter burn: The number of burn-in iterations
            :type burn: int

            :parameter every: The rate at which to save points. Saves every Nth iteration
            :type every: int

            :parameter start: The value at which to start the chains
            :type start: dict of str, float

            :parameter n_chains: The number of chains to run in parallel
            :type n_chains: int

            :parameter prior_only: Whether to run the chain on just the prior distributions.
            :type prior_only: bool

            :parameter seed: The random seed for the chains
            :type seed: int
        """
        if seed is not None:
            np.random.seed(seed)

        if n_chains == -1:
            logging.info('\nRunning Chain\nTotal: {}\nBurn: {}\nEvery: {}'
                         '\nStart: {}\nPrior Only: {}\nSeed: {}'.format(total, burn, every,
                                                                        start, prior_only, seed))
            try:
                data = self._run_chain(np.random.randint(2 ** 32 - 1), total, burn, every, start, prior_only)
            except Exception as ex:
                traceback.print_exc()
                logging.error('\nException Caught While Running Chain\n{}'.format(traceback.format_exc()))
                raise ex

            self.chains, self.log_prob, self.acceptance_rate = data
            for key, value in self.chains.items():
                self.chains[key] = [value]

            self.log_prob = [self.log_prob]
        else:
            logging.info('\nRunning {} Chains in Parallel\nTotal: {}\nBurn: {}\nEvery: {}'
                         '\nStart: {}\nPrior Only: {}\nSeed: {}'.format(n_chains, total, burn, every,
                                                                        start, prior_only, seed))

            # generate a set of seeds to pass to each chain
            seeds = np.random.randint(2 ** 32 - 1, size=n_chains)
            processes = []
            vals = []
            Q = multiprocessing.Queue()  # Create a common queue to transfer the results

            for s in seeds:
                # Start N different processes each running the chain with a different seed
                # Each process needs a different seed or all the chains will run exactly alike

                p = multiprocessing.Process(target=self._run_chain_multi,
                                            args=(s, total, burn, every, start, prior_only, Q))
                processes.append(p)
                p.start()
            try:
                for _ in processes:  # Get all the results from each process. (Results have been pushed to the queue)
                    val = Q.get()
                    if isinstance(val, Exception):
                        raise val
                    vals.append(val)
            finally:
                for process in processes:  # Join all the processes
                    process.join()

            chains, log_prob, acceptance_rate = list(zip(*vals))
            tmp_chains = {key: [] for key in chains[0].keys()}
            for chain in chains:
                for key, value in tmp_chains.items():
                    value.append(chain[key])
            self.chains = tmp_chains
            self.acceptance_rate = np.mean(acceptance_rate)
            self.log_prob = np.array(log_prob)

    def get_chains(self, flattened=False, scaled=True):
        """
            Get the chains generated from run_chains()

            Each input variable will have a list of N numpy arrays where N is the number of chains run (controlled by
            n_chains in run_chains())

            :returns: A dictionary mapping input names to lists of numpy arrays containing the values generated from
                      running the MCMC algorithm
            :rtype: dict of (str, list of numpy array of float)
        """

        def lin_scaling(value, low, high, unscaled_low, unscaled_high):
            return (unscaled_high - unscaled_low) * (value - low) / (high - low) + unscaled_low

        def log_scaling(value, low, high, unscaled_low, unscaled_high):
            #return lin_scaling(np.log10(value), np.log10(low), np.log10(high), unscaled_low, unscaled_high)
            return lin_scaling(value, low, high, np.log10(unscaled_low), np.log10(unscaled_high))

        if flattened:
            val = {key: np.array(value).flatten() for key, value in self.chains.items()}
        else:
            val = {key: np.array(value) for key, value in self.chains.items()}

        if scaled is False:
            return val
        else:
            ret = {}
            for key, value in val.items():
                if key in self.inputs:
                    current_input = self.inputs[key]
                    scaling = current_input.scaling

                    if isinstance(scaling, str):
                        if current_input.unscaled_low is not None and current_input.unscaled_high is not None:
                            low, high, unscaled_low, unscaled_high, scaling = current_input.low, current_input.high, current_input.unscaled_low, current_input.unscaled_high, current_input.scaling
                            if scaling.lower()[:3] == 'lin':
                                ret[key] = lin_scaling(value, low, high, unscaled_low, unscaled_high)
                            elif scaling.lower()[:3] == 'log':
                                ret[key] = log_scaling(value, low, high, unscaled_low, unscaled_high)
                            else:
                                raise Exception('Unscaled Upper and Lower Bound must be specified.'
                                                'Upper Bound was {}. Lower Bound was {}'.format(unscaled_high, unscaled_low))
                    else:
                        ret[key] = value
                else:
                    ret[key] = value
            return ret

    def get_diagnostics(self, n_split=2, scaled=True):
        """
            Returns a set of diagnostics on the chains generated from run_chains()

            :parameter n_split: The number of sub-chains to divide each chain into
            :type n_split: int

            :returns: Dictionary with diagnostics:

                - ``rhat``: Measures convergence of chains. Closer to 1 is better. Should be less than 1.1
                - ``n_eff``: Measures the effective sample size. Higher is better. Should be greater that 10k
                - ``var_hat``: Estimates the variance of the samples.
                - ``autocorrelation``: Estimates the autocorrelation at lag t. Closer to 0 for low t is better.
                - ``mean``: The mean of the samples
                - ``std``: The standard deviation of the samples

            :rtype: dict of str
        """
        dt_diagnostics = {'acceptance_rate': self.acceptance_rate}
        #unscaled_chains = self.get_chains(scaled=scaled)
        unscaled_chains = self.get_chains(scaled=False)
        for key in unscaled_chains.keys():
            chains_ = list(map(functools.partial(mcmc_diagnostics.split, n_splits=n_split), unscaled_chains[key]))
            current_chains = []
            for chain_ in chains_:
                for sub_chain in chain_:
                    current_chains.append(sub_chain)
            dt_diagnostics[key] = {'r_hat'          : mcmc_diagnostics.r_hat(current_chains),
                                   'n_eff'          : mcmc_diagnostics.n_eff(current_chains),
                                   'var_hat'        : mcmc_diagnostics.var_hat(current_chains),
                                   'autocorrelation': {i: mcmc_diagnostics.rho_hat(current_chains, i) for i in
                                                       range(1, len(current_chains[0]))},
                                   'mean'           : np.mean(unscaled_chains[key]),
                                   'std'            : np.std(unscaled_chains[key]),
                                   'mode'           : mcmc_diagnostics.mode(current_chains)}

        return dt_diagnostics

    def diagnostics_string(self, n_split=2, autocorr=False, scaled=True):
        """
            Returns a set of diagnostics on the chains generated from run_chains()

            :parameter n_split: The number of sub-chains to divide each chain into
            :type n_split: int

            :parameter autocorr: Whether to include auto-correlation diagnostics
            :type autocorr: bool

            :returns: formated string of  diagnostics:

                - ``rhat``: Measures convergence of chains. Closer to 1 is better. Should be less than 1.1
                - ``n_eff``: Measures the effective sample size. Higher is better. Should be greater that 10k
                - ``var_hat``: Estimates the variance of the samples.
                - ``autocorrelation``: Estimates the autocorrelation at lag t. Closer to 0 for low t is better.
                - ``mean``: The mean of the samples
                - ``std``: The standard deviation of the samples

            :rtype: str
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            diags = self.get_diagnostics(n_split=n_split, scaled=scaled)

            strs = ['acceptance_rate:{}'.format(int(diags['acceptance_rate']*10**13)/10**13)]
            for key, value in diags.items():
                if key != 'acceptance_rate':
                    strs.append('{}:'.format(key))
                    for k, v in value.items():
                        if k != 'autocorrelation':
                            strs.append('  {}:{}'.format(k, int(v*10**13)/10**13))
                        elif autocorr:
                            strs.append('  {}:'.format(k))
                            for k1, v1 in v.items():
                                strs.append('    {}:{}'.format(k1, int(v1*10**13)/10**13))

            return '\n'.join(strs)

    def get_residuals(self, scaled=True):
        """
            Returns the normalized residuals of the MCMC draws run through the surrogate models

            :returns: Normalized residuals
            :rtype: dict of list of float
        """

        def normalize(chains, output):
            residuals = []
            ls_input_vals = [chains[ipt] for ipt in output.inputs]
            for x_hat in zip(*ls_input_vals):
                x_hat = np.array(x_hat).T
                predicted = output.surrogate_model.predict(x_hat)
                residuals.append(
                    ((output.observed_value - predicted) / output.observed_std).flatten())

            return residuals

        return {key: normalize(self.get_chains(scaled=scaled), self.outputs[key]) for key in self.outputs.keys()}

    def trace_plot(self, input_name, ax=None, scaled=True):
        """
            Plots MCMC draws at each iteration number.

            :param input_name: Which input to plot
            :type input_name: str

            :param ax: The axes to plot to
            :type ax: matplotlib.axes.Axes
        """

        if ax is None:
            ax = plt
        for i, chain in enumerate(self.get_chains(scaled=scaled)[input_name]):
            ax.plot(i * len(chain) + np.arange(len(chain)), chain, linewidth=.1)
        if ax is plt:
            ax.show()

    def autocorr_plot(self, input_name, N=50, n_split=2, ax=None, scaled=True):
        """
            Plots auto-correlation values at given lag values.

            :param input_name: Which input to plot
            :type input_name: str

            :param N: The maximum lag value to plot
            :type N: int

            :param n_split: The number of chains to split each chain into when calculation auto-correlation
            :type n_split: int

            :param ax: The axes to plot to
            :type ax: matplotlib.axes.Axes
        """

        if ax is None:
            ax = plt
        autocorr = self.get_diagnostics(n_split, scaled=scaled)[input_name]['autocorrelation']
        X, Y = zip(*autocorr.items())
        ax.stem(X[:N], Y[:N], markerfmt='.')
        if ax is plt:
            ax.show()

    def residuals_plot(self, output_name, bins=10, ax=None, scaled=True):
        """
            Plots a histogram plot of the normalized residuals.

            :param output_name: Which output to plot
            :type output_name: str

            :param bins: The number of histogram bins to create
            :type bins: int

            :param ax: The axes to plot to
            :type ax: matplotlib.axes.Axes
        """

        if ax is None:
            fig, ax_tmp = plt.subplots(1, 1)
        else:
            ax_tmp = ax

        ax_tmp.axhline(3, color='xkcd:black', linestyle='--')
        ax_tmp.axhline(-3, color='xkcd:black', linestyle='--')
        for i, res in enumerate(self.get_residuals(scaled=scaled)[output_name]):
            ax_tmp.plot(i * len(res) + np.arange(len(res)), res, '.')

        divider = mpl_toolkits.axes_grid1.make_axes_locatable(ax_tmp)
        ax_histogram = divider.append_axes('right', size='10%', pad='2%', sharey=ax_tmp)

        x_norm_pdf = np.linspace(-3, 3, 1000)
        y_norm_pdf = sts.norm().pdf(x_norm_pdf)

        ax_histogram.plot(y_norm_pdf, x_norm_pdf, color='xkcd:red')
        for res in self.get_residuals(scaled=scaled)[output_name]:
            ax_histogram.hist(res, bins=bins, alpha=.5, orientation='horizontal', density=True)
        if ax is None:
            plt.show()

    def posterior_predictive_plot(self, output_name, bins=10, ax=None):
        """
            Plots a histogram plot of the posterior predictive distribution for a given output.

            :param output_name: Which output to plot
            :type output_name: str

            :param bins: The number of histogram bins to create
            :type bins: int

            :param ax: The axes to plot to
            :type ax: matplotlib.axes.Axes
        """

        def predict(chains, output):
            ls_input_vals = [chains[ipt] for ipt in output.inputs]
            for x_hat in zip(*ls_input_vals):
                x_hat = np.array(x_hat).T
                pred = output.surrogate_model.predict(x_hat)
                yield pred

        if ax is None:
            ax = plt
        opt = self.outputs[output_name]
        ax.axvline(opt.observed_value, color='xkcd:black', linestyle='--')
        for prediction in predict(self.get_chains(scaled=False), opt):
            y_tilde = np.random.normal(prediction, opt.observed_std)
            ax.hist(y_tilde, bins=bins, alpha=.5, density=True)
        if ax is plt:
            ax.show()

    def log_posterior_plot(self, ax=None):
        """
            Plots log posterior probability of draws at each iteration number.

            :param ax: The axes to plot to
            :type ax: matplotlib.axes.Axes
        """
        if ax is None:
            ax = plt
        for i, ls in enumerate(self.log_prob):
            plt.plot(i * len(ls) + np.arange(len(ls)), ls, linewidth=.1)
        if ax is plt:
            ax.show()

    def histogram_plot(self, input_name, bins=10, density=True, alpha=.5, ax=None, scaled=True):
        """
            Plots a histogram plot of the posterior distribution for a given input.

            :param input_name: Which output to plot
            :type input_name: str

            :param bins: The number of histogram bins to create
            :type bins: int

            :param density: Whether to normalize the histogram such that it integrates to 1
            :type density: bool

            :param alpha: How opaque to make the plot. 0 is not visible. 1 is totally opaque.
            :type alpha: float

            :param ax: The axes to plot to
            :type ax: matplotlib.axes.Axes
        """

        if ax is None:
            ax = plt
        ax.hist(self.get_chains(flattened=True, scaled=scaled)[input_name], bins=bins, density=density, alpha=alpha)
        if ax is plt:
            ax.show()

    def _run_chain_multi(self, seed, total, burn, every, start, prior_only, queue):
        """
            A helper method for calling _run_chain with a queue

            Gets the value from _run_chain() and pushes it to the queue
        """
        try:
            val = self._run_chain(seed, total, burn, every, start, prior_only)
            queue.put(val)
        except Exception as ex:
            traceback.print_exc()
            logging.error('\nException Caught While Running Chain\n{}'.format(traceback.format_exc()))
            queue.put(ex)


class DefaultMCMC(MCMC):

    def _run_chain(self, seed, total, burn, every, start, prior_only=False):
        """
            Run a single MCMC chain

            Number of iterations = total*every + burn

            Iterates by asking _metropolis for a point. Only stores every Nth point where N is the 'every' parameter.

            :parameter seed: The seed for the chain
            :type seed: int

            :parameter total: The total number of sample points to return
            :type total: int

            :parameter burn: The number of burn-in iterations
            :type burn: int

            :parameter every: The rate at which to save points. Saves every Nth iteration
            :type every: int

            :parameter start: The value at which to start the chain
            :type start: dict of str, float

            :returns: **MCMC chain**: The values saved through the entire MCMC process
                      **acceptance rate**: The ratio between the number of accepted new points and total points
            :rtype: dict of str, numpy array of float
        """
        rng = np.random.RandomState(seed=seed)

        np_log_prob = np.zeros(total)

        start = self._init_start(start, rng)

        chain = {key: np.zeros(total) for key in start.keys()}

        accepted_total = 0
        i = 0
        j = 0
        current_point = start
        log_prob = None
        logging.info('Burning')
        print('Burning')
        while i < total:
            new_point, log_prob, accepted = self._metropolis(current_point, rng, prior_only, log_prob_current=log_prob)

            if j == burn + 1:
                logging.info('Sampling Start')
                print('Sampling Start')
                sys.stdout.flush()
            if j > burn and j % every == 0:
                for key in start.keys():
                    chain[key][i] = new_point[key]
                np_log_prob[i] = log_prob
                i += 1
                if i % (total / 10) == 0:
                    print("Sampling: {0:g}%".format((float(i) / float(total)) * 100))
                    sys.stdout.flush()
            j += 1

            accepted_total += accepted
            current_point = new_point

        logging.info("Sampling Done")
        print("Sampling Done")
        return chain, np_log_prob, float(accepted_total) / float(j)

    def _init_start(self, start, rng):
        if start is None:
            return {key: rng.uniform(self.inputs[key].low, self.inputs[key].high) for key in
                    self.inputs.keys()}
        else:
            return start

    def _metropolis(self, current_point, rng, prior_only=False, log_prob_current=None):
        """
            Accepts a new point or Rejects new point and keeps old point

            Proposes a new point and compares the ratio of the log probabilities of the new point and the old point to a
            uniformly drawn number. If the ratio is higher, the new point is accepted and returned. If the ratio is
            lower, the new point is rejected and the old point is returned.

            :parameter current_point: The current point of the markov chain
            :type current_point: dict of str, float

            :returns: **new point**: The new point if accepted or the old point if rejected
                      **accepted**: Whether the new point was accepted
            :rtype: (dict of str, float, bool)
        """
        proposed_point = self._propose(current_point, rng)

        log_prob_proposed = self._posterior_log_prob(proposed_point, prior_only)
        if log_prob_current is None:
            log_prob_current = self._posterior_log_prob(current_point, prior_only)

        f_alpha = np.exp(log_prob_proposed - log_prob_current)

        u = rng.uniform()
        if u < f_alpha:
            logging.debug('\nAccepted Point\nCurrent Point: {}\nNew Point: {}\n'
                          'Current Log Prob: {}\nNew Log Prob:'
                          '{}\nf alpha:{}\nu: {}'.format(current_point,
                                                         proposed_point,
                                                         log_prob_current,
                                                         log_prob_proposed,
                                                         f_alpha,
                                                         u))
            return proposed_point, log_prob_proposed, True
        else:
            logging.debug('\nRejected Point\nCurrent Point: {}\nNew Point: {}\n'
                          'Current Log Prob: {}\nNew Log Prob:'
                          '{}\nf alpha:{}\nu: {}'.format(current_point,
                                                         proposed_point,
                                                         log_prob_current,
                                                         log_prob_proposed,
                                                         f_alpha,
                                                         u))
            return current_point, log_prob_current, False

    def _propose(self, current_point, rng):
        """
            Propose a new point which is conditioned on the current point

            :parameter current_point: The current point which is the pre-state for the proposed point
            :type current_point: dict of (str, float)

            :returns: A new point whose distribution depends on the given current point
            :rtype: dict of (str, float)
        """
        return {key: rng.normal(current_point[key], self.inputs[key].proposal_sigma) for key in
                current_point.keys()}

    def _posterior_log_prob(self, current_point, prior_only=False):
        """
            Gets the log probability for a given point

            :parameter current_point: The given point to get the log probability for
            :type current_point: dict of (str, float)

            :returns: The log probability of the given point
        """

        num_outputs = len(self.outputs.keys())
        likelihood = 0.0
        y_vec = np.zeros(num_outputs)
        z_vec = np.zeros(num_outputs)
        sigma_y = np.zeros(num_outputs)
        sigma_z = np.zeros(num_outputs)

        if not prior_only:
            for idx, output in enumerate(self.outputs.values()):
                ls_input_vals = [current_point[ipt] for ipt in output.inputs]
                pred_result, std_result = output.surrogate_model.predict(np.array([ls_input_vals]), return_std=True)
                z_vec[idx] = pred_result.item() if hasattr(pred_result, 'item') else pred_result[0]
                z_std = std_result.item() if hasattr(std_result, 'item') else std_result[0]
                y_vec[idx] = output.observed_value
                sigma_y[idx] = output.observed_std
                if z_std is not None:
                    sigma_z[idx] = z_std.item() if hasattr(z_std, 'item') else z_std

            likelihood = sts.multivariate_normal(y_vec, np.diag(sigma_y ** 2) + np.diag(sigma_z ** 2)).logpdf(z_vec)

        def bounding_prior(input, x):
            return float('-inf') if x < input.low or x > input.high else input.prior(x)

        ls_prior = [bounding_prior(self.inputs[key], current_point[key]) for key in current_point.keys()]

        log_prob = likelihood + np.sum(ls_prior)
        logging.debug('\nCurrent Point: {}\nLog Prob: {}'.format(current_point, log_prob))
        return log_prob


class DiscrepancyMCMC(DefaultMCMC):

    def __init__(self):
        super(DiscrepancyMCMC, self).__init__()
        self.num_quant = None
        self.quantities = None
        self.tau_proposal_sigma = .05
        self.rho_proposal_sigma = .02
        self.tau_prior_alpha = 0.0
        self.tau_prior_beta = 0.0

    def _init_start(self, start, rng):
        """
            Start values for the chain. Discrepancy covariance variables are added here.
            Tau is the discrepancy variance for each quantity of interest, and rho is the
            correlation coefficient between the two quantities of interest.

            :parameter start: The value at which to start the chain
            :type start: dict of str, float
        """
        quantities = []
        for output_name, output in self.outputs.items():
            if output.quantity not in quantities:
                quantities.append(output.quantity)

        self.num_quant = len(quantities)
        self.quantities = quantities

        if start is None:
            initial = {key: rng.uniform(self.inputs[key].low, self.inputs[key].high) for key in
                       self.inputs.keys()}

            tau = ['tau_{}'.format(i) for i in self.quantities]

            for t1 in tau:
                for t2 in tau:
                    first = t1.lstrip('tau_')
                    second = t2.lstrip('tau_')
                    if first != second and 'rho_{}_{}'.format(second, first) not in initial:
                        initial['rho_{}_{}'.format(first, second)] \
                            = 0.0 if self.rho_proposal_sigma == 0.0 else rng.uniform(-1, 1)

            for t in tau:
                initial[t] = sts.invgamma(1, 1).ppf(rng.uniform())
                #initial[t] = max(1.0e-6, rng.uniform(0.0, 3.0*self.tau_proposal_sigma)
        else:
            initial = deepcopy(start)

        return initial

    def _propose(self, current_point, rng):
        """
            Propose a new point which is conditioned on the current point

            :parameter current_point: The current point which is the pre-state for the proposed point
            :type current_point: dict of (str, float)

            :returns: A new point whose distribution depends on the given current point
            :rtype: dict of (str, float)
        """
        proposed_point = {
            key: rng.normal(current_point[key], self.inputs[key].proposal_sigma) if key in self.inputs else 0.0
            for key in current_point.keys()}

        for key in filter(lambda x: x.startswith('tau_'), current_point.keys()):
            proposed_point[key] = max(1e-6, rng.normal(current_point[key], self.tau_proposal_sigma))

        rps = self.rho_proposal_sigma
        for key in filter(lambda x: x.startswith('rho_'), current_point.keys()):
            proposed_point[key] \
                = 0.0 if rps == 0.0 else rng.normal(current_point[key], rps)

        return proposed_point

    def _posterior_log_prob(self, current_point, prior_only=False):
        """
            Gets the log probability for a given point

            :parameter current_point: The given point to get the log probability for
            :type current_point: dict of (str, float)

            :returns: The log probability of the given point
        """
        num_outputs = len(self.outputs.keys())
        likelihood = 0.0
        tau_prob = np.zeros(self.num_quant)

        if not prior_only:
            z_vec = np.zeros(num_outputs)
            y_vec = np.zeros(num_outputs)
            sigma_delta = np.zeros((num_outputs, num_outputs))
            sigma_y = np.zeros(num_outputs)
            sigma_z = np.zeros(num_outputs)

            for idx, output in enumerate(self.outputs.values()):
                ls_input_vals = [current_point[ipt] for ipt in output.inputs]
                pred_result, std_result = output.surrogate_model.predict(np.array([ls_input_vals]), return_std=True)
                z_vec[idx] = pred_result.item() if hasattr(pred_result, 'item') else pred_result
                z_std = std_result.item() if hasattr(std_result, 'item') else std_result
                y_vec[idx] = output.observed_value
                sigma_y[idx] = output.observed_std ** 2
                sigma_delta[idx, idx] = current_point['tau_{}'.format(output.quantity)] ** 2
                if z_std is not None:
                    sigma_z[idx] = z_std ** 2

                for jdx, output_alt in enumerate(self.outputs.values()):
                    if output.event == output_alt.event:
                        if idx != jdx:
                            rho1 = 'rho_{}_{}'.format(output.quantity, output_alt.quantity)
                            rho2 = 'rho_{}_{}'.format(output_alt.quantity, output.quantity)
                            if rho1 in current_point:
                                if not -1.0 <= current_point[rho1] <= 1.0:
                                    return float('-inf')
                                sigma_delta[idx, jdx] = current_point[rho1] * \
                                                        current_point['tau_{}'.format(output.quantity)] * \
                                                        current_point['tau_{}'.format(output_alt.quantity)]
                            elif rho2 in current_point:
                                if not -1.0 <= current_point[rho2] <= 1.0:
                                    return float('-inf')
                                sigma_delta[idx, jdx] = current_point[rho2] * \
                                                        current_point['tau_{}'.format(output.quantity)] * \
                                                        current_point['tau_{}'.format(output_alt.quantity)]

            try:
                likelihood = sts.multivariate_normal(mean=y_vec,
                                                     cov=np.linalg.multi_dot([np.diag(z_vec),
                                                                              sigma_delta,
                                                                              np.diag(z_vec)]) +
                                                         np.diag(sigma_y) + np.diag(sigma_z)).logpdf(z_vec)
            except (ValueError, np.linalg.LinAlgError):
                return float('-inf')

        for idx, quantity in enumerate(self.quantities):
            tau_prob[idx] = self._square_log_invgamma_proportional(current_point['tau_{}'.format(quantity)],
                                                                   self.tau_prior_alpha, self.tau_prior_beta)

        def bounding_prior(input, x):
            return float('-inf') if x < input.low or x > input.high else input.prior(x)

        ls_prior = [bounding_prior(self.inputs[key], current_point[key]) if key in self.inputs else 0 for key in
                    current_point.keys()]

        log_prob = likelihood + np.sum(ls_prior) + np.sum(tau_prob)

        logging.debug('\nCurrent Point: {}\nLog Prob: {}'.format(current_point, log_prob))

        return log_prob

    @staticmethod
    def _square_log_invgamma_proportional(x, alpha, beta):
        if x < 0:
            return float('-inf')
        return -(2 * alpha + 1) * np.log(x) - beta / (x ** 2)


class OutputVariable:
    """
        Represents an observed output
    """

    def __init__(self, event, quantity, surrogate_model, observed_value, observed_std, inputs):
        """
            :parameter name: The unique name of the output
            :type name: str

            :parameter surrogate_model: The surrogate model that represents the mapping of input values to output
                                        values. This model must be have a predict method that takes a numpy array and
                                        returns a numpy array (sklearn's fit/predict paradigm)

            :parameter observed_value: The observed experimental value
            :type observed_value: float

            :parameter observed_std: The error bound on the observed value
            :type observed_std: float

            :parameter inputs: The ordered names of the input variables that map to the output
            :type inputs: list of str
        """
        self.name = event + '_' + quantity
        self.surrogate_model = surrogate_model
        self.observed_value = observed_value
        self.observed_std = observed_std
        self.inputs = inputs
        self.event = event
        self.quantity = quantity


class InputVariable:
    """
        Represents an unobserved input
    """

    def __init__(self, name, low, high, proposal_sigma, prior=None, unscaled_low=None, unscaled_high=None, scaling=None):
        """

            :parameter name: The unique name of the input
            :type name: str

            :parameter low: The lower bound on the value this variable can take
            :type low: float

            :parameter high: The upper bound on the value this variable can take
            :type high: float

            :parameter proposal_sigma: The standard deviation of the proposal distribution for this variable
            :type proposal_sigma: float

            :parameter prior: The prior distribution of the variable.
                              If None, a uniform distribution over the whole range is used
            :type prior: function
        """
        self.name = name
        self.low = low
        self.high = high
        self.proposal_sigma = proposal_sigma
        if prior is None:
            self.prior = lambda x: 1.0
        else:
            self.prior = prior

        self.unscaled_low = unscaled_low
        self.unscaled_high = unscaled_high
        self.scaling = scaling


if __name__ == "__main__":
    from sklearn.gaussian_process import GaussianProcessRegressor

    logging.basicConfig(filename='foobar.log', level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s\n')

    X = np.random.randn(100, 2) * 10
    Y = np.sqrt(np.power(X, 2).sum(axis=1).reshape(-1, 1))

    mdl = GaussianProcessRegressor().fit(X, Y)

    mdl2 = GaussianProcessRegressor().fit(X + 1.0, Y)

    exp1 = DefaultMCMC()

    exp1.tau_prior_beta = 0.0
    exp1.tau_prior_alpha = 0.0
    exp1.tau_proposal_sigma = 1.0
    exp1.rho_proposal_sigma = .5

    exp1.add_input('foo', -10, 10, 1.0, sts.norm.logpdf, unscaled_low=0, unscaled_high=1, scaling='log')

    exp1.add_input('bar', -10, 10, 1.0, sts.norm.logpdf, 1, 2)

    exp1.add_output('ax', 'height', mdl, 3, .4, ['foo', 'bar'])
    exp1.add_output('ay', 'height', mdl2, 3, .3, ['foo', 'bar'])
    exp1.add_output('ax', 'width', mdl, 3, .4, ['foo', 'bar'])
    exp1.add_output('ay', 'width', mdl2, 3, .3, ['foo', 'bar'])

    # exp1.run_chain(total=1000, burn=2000, every=50, n_chains=10, prior_only=True)
    # prior_points = exp1.get_chains(flattened=True)

    exp1.run_chain(total=1000, burn=200, every=5, n_chains=2)
    post_points = exp1.get_chains(flattened=True)

    print(exp1.diagnostics_string())
    logging.info('Diagnostics:\n' + exp1.diagnostics_string())

    # for ipt in exp1.get_chains().keys():
    #     fig, ax = plt.subplots(1, 1)
    #     exp1.trace_plot(ipt, ax=ax)
    #     fig.savefig('trace_plot_{}.png'.format(ipt))
    #     plt.close(fig)
    #
    #     fig, ax = plt.subplots(1, 1)
    #     exp1.autocorr_plot(ipt, ax=ax)
    #     fig.savefig('autocorr_plot_{}.png'.format(ipt))
    #     plt.close(fig)
    #
    #     fig, ax = plt.subplots(1, 1)
    #     exp1.histogram_plot(ipt, bins=100, density=True, ax=ax)
    #     fig.savefig('autocorr_plot_{}.png'.format(ipt))
    #     plt.close(fig)
    #
    # for opt in exp1.outputs.values():
    #     fig, ax = plt.subplots(1, 1)
    #     exp1.posterior_predictive_plot(opt.name, ax=ax)
    #     fig.savefig('posterior_predictive_plot_{}.png'.format(opt.name))
    #     plt.close(fig)
    #
    #     fig, ax = plt.subplots(1, 1)
    #     exp1.residuals_plot(opt.name, ax=ax)
    #     fig.savefig('residuals_plot_{}.png'.format(opt.name))
    #     plt.close(fig)
    #
    # fig, ax = plt.subplots(1, 2)
    # exp1.log_posterior_plot(ax=ax[0])
    exp1.histogram_plot('foo')
    exp1.histogram_plot('foo', scaled=False)
    # fig.savefig('log_posterior_plot1.png')
    # plt.close(fig)

    # plots.contour_plot(plt, post_points['foo'], post_points['bar'])
    # plt.show()
    # plots.scatter_plot(plt, post_points['foo'], post_points['bar'], 100)
    # plt.show()



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from itertools import count

import numpy as np


def get_chains_size(chains):
    chains = np.array(chains)
    chains_shape = chains.shape
    if len(chains_shape) != 2:
        raise Exception("Given chains are not all the same length")

    return chains_shape


def within_variance(chains):
    # mean of variances
    return np.mean([np.var(chain, ddof=1) for chain in chains])


def between_variance(chains):
    # variance of means
    return np.var([np.mean(chain) for chain in chains], ddof=1) * get_chains_size(chains)[1]


def var_hat(chains):
    n = len(chains[0])

    return ((n - 1.) / n) * within_variance(chains) + (between_variance(chains) / n)


def r_hat(chains):
    return np.sqrt(var_hat(chains) / within_variance(chains))


def variogram(chains, t):
    return np.mean([np.mean(np.power(chain[:-t] - chain[t:], 2)) for chain in chains])


def rho_hat(chains, t, prev_var_hat=None):
    if prev_var_hat is None:
        return 1. - variogram(chains, t) / (2 * var_hat(chains))
    else:
        return 1. - variogram(chains, t) / (2 * prev_var_hat)


def n_eff(chains):
    m, n = get_chains_size(chains)
    prev_var_hat = var_hat(chains)

    s = 0.0

    it1 = (rho_hat(chains, t, prev_var_hat) for t in count(1))
    for x in it1:
        val = float(x + next(it1))
        if val >= 0:
            s += val
        else:
            break

    return m * n / (1 + 2 * s)


def split(chain, n_splits=2):
    def chunk():
        n = len(chain) // n_splits
        for i in range(0, len(chain), n):
            yield chain[i:i + n]

    return np.array(list(chunk()))


def mode(chains, bins=100):
    height, bins = np.histogram(chains, bins=bins)
    max_bin = np.argmax(height)
    return np.mean([bins[max_bin], bins[max_bin+1]])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib.pyplot import cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from sklearn.feature_selection import mutual_info_regression, f_regression
from sklearn.linear_model import lars_path
from sklearn.preprocessing import PolynomialFeatures

from trata.sampler import LatinHyperCubeSampler
from ibis.pce_model import PolynomialChaosExpansionModel


def _variance_network_plot(ax,
                           feature_data,
                           response_data,
                           feature_names,
                           response_names,
                           score_function,
                           method_label,
                           degree=2,
                           max_size=20.0,
                           alpha=.5,
                           label_size=12,
                           **kwargs):
    """
    Create network plots with consistent scaling across all degree subplots.
    
    ARCHITECTURAL NOTE:
    This function supports two types of sensitivity methods:
    - Type A (interaction-based): Methods that work directly on pre-computed interactions
      (e.g., _f_score, _mutual_info_score)
    - Type B (original feature-based): Methods that need raw features to build their own basis
      (e.g., _pce_score which builds polynomial basis internally)
    
    For Type B methods, the score function should extract original features from the
    interaction data using: feature_data[:, powers.sum(axis=-1) == 1]
    """
    assert degree >= 2

    # Create interaction terms for plotting structure
    feature_interaction_data, feature_interaction_names, powers = _make_interactions(
        feature_data,
        feature_names,
        degree=degree,
        interaction_only=True
    )
    feature_interaction_names = np.array(feature_interaction_names)
    powers_degree = powers.sum(axis=1)

    # Reshape axis for case with one output gives a 1D array; force to 2D
    if len(ax.shape) == 1:
        ax = ax.reshape(-1, 1)

    for response_column_data, response_column_name, plot_column in zip(
        response_data.T, response_names, ax.T
    ):
        # Compute scores for all interaction terms
        scores = score_function(
            feature_interaction_data, 
            response_column_data, 
            powers=powers, 
            **kwargs
        )

        # Extract node weights (degree 1 terms)
        node_weights = scores[powers_degree == 1]
        node_names = np.array(feature_names)

        # Calculate global size_factor across all degrees for consistency
        all_scores = [node_weights]
        for deg in range(2, degree + 1):
            all_scores.append(scores[powers_degree == deg])
        
        global_max = max(s.max() for s in all_scores if len(s) > 0)
        size_factor = max_size / global_max if global_max > 0 else 1.0
        
        # Calculate node sizes once (they're the same across all subplots)
        node_sizes = size_factor * node_weights
        node_labels = {
            name: f"{name}\n{weight:.2e}" 
            for name, weight in zip(node_names, node_weights)
        }

        # Plot each degree in separate subplot
        for degree_to_plot, plot_row in zip(range(2, degree + 1), plot_column):
            edge_weights = scores[powers_degree == degree_to_plot]
            edge_sizes = size_factor * edge_weights

            # Use nonzero() instead of astype('bool') for clarity
            edge_list = [
                tuple(node_names[p.nonzero()[0]])
                for p in powers[powers_degree == degree_to_plot]
            ]
            edge_names = feature_interaction_names[powers_degree == degree_to_plot]
            edge_labels = {
                names: f"{int_name}\n{weight:.2e}"
                for names, int_name, weight in zip(edge_list, edge_names, edge_weights)
            }

            if degree_to_plot == 2:
                # Standard graph for pairwise interactions
                graph = nx.complete_graph(node_names)
                pos = nx.circular_layout(graph, scale=1.0)

                nx.draw_networkx_nodes(
                    graph, pos,
                    node_size=node_sizes ** 2,
                    node_color='xkcd:red',
                    linewidths=1.0,
                    alpha=alpha,
                    ax=plot_row
                )

                nx.draw_networkx_labels(
                    graph, pos,
                    labels=node_labels,
                    font_size=label_size,
                    ax=plot_row
                )

                nx.draw_networkx_edges(
                    graph, pos,
                    width=edge_sizes,
                    edge_color='xkcd:green',
                    alpha=alpha,
                    ax=plot_row
                )

                nx.draw_networkx_edge_labels(
                    graph, pos,
                    edge_labels=edge_labels,
                    font_size=label_size,
                    label_pos=.6,
                    ax=plot_row
                )

            else:
                # Hypergraph for higher-order interactions
                graph = nx.Graph()
                graph.add_nodes_from(node_names)
                graph.add_nodes_from(edge_list)
                
                for edge, weight in zip(edge_list, edge_weights):
                    for node in edge:
                        graph.add_edge(node, edge, weight=weight)

                hyper_edge_weights = [size_factor * el['weight'] for el in graph.edges.values()]
                pos = nx.bipartite_layout(graph, edge_list, align='horizontal')

                nx.draw_networkx_nodes(
                    graph, pos,
                    nodelist=node_names,
                    node_size=node_sizes ** 2,
                    node_color='xkcd:red',
                    linewidths=1.0,
                    alpha=alpha,
                    ax=plot_row
                )

                nx.draw_networkx_nodes(
                    graph, pos,
                    nodelist=edge_list,
                    node_size=edge_sizes ** 2,
                    node_color='xkcd:green',
                    linewidths=1.0,
                    alpha=alpha,
                    ax=plot_row
                )

                nx.draw_networkx_labels(
                    graph, pos,
                    labels=node_labels,
                    font_size=label_size,
                    ax=plot_row
                )

                nx.draw_networkx_labels(
                    graph, pos,
                    labels=edge_labels,
                    font_size=label_size,
                    ax=plot_row
                )

                nx.draw_networkx_edges(
                    graph, pos,
                    edge_color='xkcd:green',
                    width=hyper_edge_weights,
                    alpha=alpha,
                    ax=plot_row
                )

            plot_row.set_title(f"'{response_column_name}' degree {degree_to_plot} ({method_label})")


def _rank_plot(ax, feature_data, response_data, feature_names, response_names, score_function,
               degree=1, interaction_only=True, **kwargs):
    """
    Create a rank plot based on a given set of data and score function.

    Ranks each feature from 1 to N based on the result of score_function.
    A grid of boxes are plotted with response names on the x-axis
    and feature names (and possibly interactions) on the y-axis.
    Each box is colored and labeled according to its calculated rank.
    Ranks are only calculated within responses, not across responses.

    Args:
        - ax (matplotlib.Axes): Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column
          is a feature; each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond
          to rows in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - score_function (function): Function which scores features based on
          importance to responses.
        - degree (int): Maximum degree of interaction
        - interaction_only (bool): Whether to only include lowest powers of
          interaction or include higher powers.
        - kwargs: Keyword arguments passed to score_function

    Raises:
        -
    """
    feature_interaction_data, feature_interaction_names, powers = \
        _make_interactions(feature_data, feature_names, degree=degree, interaction_only=interaction_only)

    num_features = feature_interaction_data.shape[1]
    num_responses = response_data.shape[1]

    if "pce_score" in str(score_function):
        scores = np.vstack([score_function(feature_interaction_data,
                                              response_column, powers=powers,
                                              **kwargs) for response_column in response_data.T])

    else:
        scores = np.vstack([score_function(feature_interaction_data,
                                              response_column,
                                              **kwargs) for response_column in response_data.T])

    order = np.argsort(-scores)
    rank_table = np.argsort(order)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    color_mesh = ax.pcolormesh(num_features - rank_table.T, edgecolors='None', lw=0.0, cmap=cm.plasma)

    ax.hlines(np.arange(1, num_features + 1), 0, num_responses, color='k')
    ax.vlines(np.arange(1, num_responses + 1), 0, num_features, color='k')

    cbar = plt.colorbar(color_mesh, cax=cax, orientation='vertical')

    for i in range(num_responses * num_features):
        idx = np.unravel_index(i, (num_responses, num_features))
        ax.text(idx[0] + .5, idx[1] + .5, rank_table[idx] + 1, ha='center', va='center')

    ax.set_xticks(np.arange(num_responses) + 0.5)
    ax.set_xticklabels(response_names, rotation=35, ha='right')
    ax.set_xlabel('Outputs', ha='center')

    ax.set_yticks(np.arange(num_features) + 0.5)
    ax.set_yticklabels(feature_interaction_names)
    ax.set_ylabel('Parameters')

    cbar_labels = np.arange(1, num_features + 1).astype('str')
    cbar_labels[-1] += ' (Least Sensitive)'
    cbar_labels[0] += ' (Most Sensitive)'

    cbar.set_ticks(num_features - np.arange(num_features))
    cbar.set_ticklabels(cbar_labels)

    ax.set_xlim([0, num_responses])
    ax.set_ylim([0, num_features])
    ax.invert_yaxis()


def _score_plot(ax, feature_data, response_data, feature_names, response_names, score_function, title,
                degree=1, interaction_only=True, y_axis_label='Score', **kwargs):
    """
    Create a set of bar plots based on a given set of data and score function for each output.

    Uses score_function to assign a score to each feature or interaction of features.
    These scores are then plotted as a bar plot, with the height of each bar corresponding to
    the calculated score.

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - score_function (function): Function which scores features based on importance to
          responses.
        - title (string): Title of plot
        - degree (int): Maximum degree of interaction
        - interaction_only (bool): Whether to only include lowest powers of interaction or
          include higher powers.
        - y_axis_label (string): Y axis label
        - kwargs: Keyword arguments passed to score_function

    Raises:
        -
    """
    feature_interaction_data, feature_interaction_names, powers = \
        _make_interactions(feature_data, feature_names, degree=degree, interaction_only=interaction_only)

    for y, y_name, axes in zip(response_data.T, response_names, ax):
        scores = score_function(feature_interaction_data, y, powers=powers, **kwargs)

        axes.bar(feature_interaction_names, scores)
        axes.tick_params(axis='x', labelrotation=70)
        axes.set_xlabel('Parameter Interaction')
        axes.set_ylabel(y_axis_label)
        axes.set_title('{} ({})'.format(title, y_name))


def _make_interactions(feature_data, feature_names, degree=2, interaction_only=False):
    """
    Create interactions between features using sklearn's PolynomialFeatures.
    
    Args:
        feature_data ([[float]]): Array-like of feature data
        feature_names ([string]): Array-like of feature names
        degree (int): Maximum degree of interaction
        interaction_only (bool): Whether to only include lowest powers of interaction
          or include higher powers.
    
    Returns:
        - Feature interaction data ([[float]]): The resulting feature data from the interactions
        - Feature interaction names ([string]): The resulting feature names from the interactions
        - Feature interaction powers ([[int]]): An array containing the powers to which each feature
                                                was raised to create the interactions.
    """
    transformer = PolynomialFeatures(
        degree=degree,
        interaction_only=interaction_only,
        include_bias=False
    )
    
    transformer.fit(feature_data)
    feature_interaction_data = transformer.transform(feature_data)
    feature_interaction_names = transformer.get_feature_names_out(feature_names)
    
    return feature_interaction_data, feature_interaction_names, transformer.powers_


def _f_score(feature_data, response_data, center=True, **kwargs):
    """
    Scores features based on an F-test of linear regression coefficients

    Each features is scored using an F-test of linear regression coefficients.
    The F statistic is calculated by looking at the difference between a linear model with
    and without the feature included.
    See sklearn.feature_selction.f_regression for more information.

    Args:
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - center (bool): Whether to center data feature_data before calculating scores
        - kwargs: Captures superfluous keyword arguments

    Returns:
        - Scores ([float]): The scores of each feature

    Raises:
        -
    """
    scores, p_values = f_regression(feature_data, response_data, center=center)
    return scores


def _f_p_score(feature_data, response_data, center=True, **kwargs):
    """
    Give the p-value for an F-test of linear regression coefficients for each feature.

    Each features is scored using an F-test of linear regression coefficients.
    The F statistic is calculated by looking at the difference between a linear model with
    and without the feature included.
    The p-value of this F statistic is returned.
    In general, a p-value < 0.05 is considered significant, meaning that particular feature
    is worth keeping.
    See sklearn.feature_selction.f_regression for more information.

    Args:
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - center (bool): Whether to center data feature_data before calculating scores
        - kwargs: Captures superfluous keyword arguments

    Returns:
        - p_values ([float]): The p-values of each feature

    Raises:
        -
    """
    scores, p_values = f_regression(feature_data, response_data, center=center)
    return p_values


def _mutual_info_score(feature_data, response_data, n_neighbors=3, **kwargs):
    """
    Scores each feature based on the amount of mutual information shared with the response.

    Each feature is scored using an estimate of mutual information between it and the response data.
    Mutual information measures the amount of information that can be obtained about the response
    by observing the particular feature.
    See sklearn.eature_selection.mutual_info_regression for more information.

    Args:
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - n_neighbors (int): How many neighboring bins to consider when estimating mutual
          information.
        - kwargs: Captures superfluous keyword arguments

    Returns:
        - Scores ([float]): The scores of each feature

    Raises:
        -
    """
    return mutual_info_regression(feature_data, response_data, n_neighbors=n_neighbors)


def _pce_score(feature_data, response_data, ranges, powers, pce_degree=1, model_degrees=1, **kwargs):
    """
    Scores features using PCE variance decomposition (Sobol indices).
    
    IMPORTANT ARCHITECTURAL NOTE:
    This function receives feature_interaction_data from _variance_network_plot,
    but PCE needs ORIGINAL features to build its orthogonal polynomial basis.
    
    The extraction works because:
    1. _make_interactions with interaction_only=True places original features first
    2. powers.sum(axis=-1) == 1 selects exactly those first n_features columns
    
    LIMITATIONS:
    - Only guaranteed to work when pce_degree == model_degrees
    - Relies on specific ordering from _make_interactions
    
    Args:
        feature_data ([[float]]): Interaction data (original features extracted internally)
        response_data ([[float]]): Response data
        ranges ([[float]]): Feature ranges for normalization
        powers ([[int]]): Power matrix from _make_interactions
        pce_degree (int): Max degree of interaction to score (should equal model_degrees)
        model_degrees (int): Max degree for PCE polynomial basis (should equal pce_degree)
        
    Returns:
        - Scores ([float]): Sobol indices ordered to match powers array
    """
    # EXTRACTION HACK: Get original features from interaction data
    # This works because _make_interactions puts original features first
    n_features = (powers.sum(axis=-1) == 1).sum()
    feature_data_to_fit = feature_data[:, :n_features]
    
    # Verify we extracted the right number of features
    assert feature_data_to_fit.shape[1] == len(ranges), \
        f"Extracted {feature_data_to_fit.shape[1]} features but ranges has {len(ranges)}"
    
    # Fit PCE model on original features
    model = PolynomialChaosExpansionModel(num_degrees=model_degrees, ranges=ranges)
    model.fit(feature_data_to_fit, response_data[:, np.newaxis])
    
    # Get Sobol indices
    # NOTE: This only matches powers order when pce_degree == model_degrees
    contributions = model.make_contributions(pce_degree).reshape(-1)
    
    # Verify output size matches expected
    expected_size = (powers.sum(axis=-1) <= pce_degree).sum()
    assert len(contributions) == expected_size, \
        f"Expected {expected_size} scores but got {len(contributions)}"
    
    return contributions


def one_at_a_time_effects(feature_data, response_data):
    """
    Calculates the elementary effects for a one-at-a-time study.

    The `feature_data` points should have been generated from trata.sampler.OneAtATimeSampler
    with `do_oat=True` or similar.
    Two elementary effects are calculated per feature resulting in 2*p total effects

    Args:
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature; each row
          is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows in feature
          data.

    Returns:
        - Elementary Effects ([[float]]): (p by 2) Numpy Array of elementary effects
    """
    m, n = feature_data.shape
    assert m == 2*n+1

    diff_features = feature_data[0] - feature_data
    diff_response = response_data[0] - response_data

    which_var = diff_features != 0

    val = diff_response[1:]/diff_features[which_var]

    return val.reshape(-1, 2).T


def morris_effects(feature_data, response_data, tol=1e-12):
    """
    Calculates the elementary effects and Morris statistics for a Morris one-at-a-time study.

    The `feature_data` points should have been generated from trata.sampler.MorrisOneAtATimeSampler
    or similar.
    The number of elementary effects calculated depends on the number of paths in the study.
    Each path produces one elementary effect per dimension.
    So for r paths, p*r total effects will be calculated.

    Args:
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature; each row
          is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows in feature
          data.
        - tol (float): Tolerance for numerical stability in parameter change detection.

    Returns:
        - dict: Contains 'elementary_effects' (r by k array), 'mu', 'mu_star', and 'sigma' arrays
    """
    # Input validation and conversion
    feature_data = np.asarray(feature_data)
    response_data = np.asarray(response_data)
    
    # Handle 1D response data
    if response_data.ndim == 1:
        response_data_1d = response_data
    else:
        response_data_1d = response_data.squeeze()

    if feature_data.shape[0] != len(response_data_1d):
        raise ValueError(f"Feature and response data must have same number of rows. "
                        f"Feature: {feature_data.shape[0]}, Response: {len(response_data_1d)}")

    n, k = feature_data.shape
    if n % (k + 1) != 0:
        raise ValueError(f"Number of samples ({n}) must be divisible by (k+1) = {k+1}")

    r = int(n / (k + 1))

    def make_effect(feature_partition, response_partition):
        diff_response = response_partition[1:] - response_partition[:-1]
        diff_features = feature_partition[1:] - feature_partition[:-1]
        which_var = np.where(np.abs(diff_features) > tol)

        # Check that exactly one parameter changes per step
        changes_per_step = np.sum(np.abs(diff_features) > tol, axis=1)
        if not np.all(changes_per_step == 1):
            raise ValueError("Invalid Morris path: multiple or no parameters changed in a step")

        val = diff_response / diff_features[which_var]
        # Re-order val
        val[which_var[1]] = val[range(k)]
        return val

    elementary_effects = np.array([make_effect(_x, _y) for _x, _y in zip(feature_data.reshape(r, k + 1, k),
                                                                    response_data_1d.reshape(r, k + 1))])

    # Calculate Morris statistics
    mu = np.mean(elementary_effects, axis=0)
    mu_star = np.mean(np.abs(elementary_effects), axis=0)
    sigma = np.std(elementary_effects, axis=0, ddof=1)

    return {
        'mu': mu,
        'mu_star': mu_star,
        'sigma': sigma,
        'elementary_effects': elementary_effects
    }


def sobol_indices(feature_data, response_data, include_second_order=False, **kwargs):
    """
    Calculates Sobol indices

    The 'feature_data' points should have been generated from
    trata.sampler.SobolIndexSampler or similar.

    Args:
        - feature_data ([[float]]): Array-like of feature data
          Each column is a feature; each row is an observation
        - response_data ([[float]]): Array-like of response data
          Rows correspond to rows in feature data. There should
          only be one column of outputs for each row of inputs.

    Returns:
        - First-order indices ([[float]]): (1 by k) Numpy array of first-order indices
        - Total-order indices ([[float]]): (1 by k) Numpy array of total-order indices
        - Second-order indices ([[float]]): (k by k) Numpy array of second-order indices
          (upper triangular)

    Raises:
        - ValueError, TypeError
    """
    r, k = feature_data.shape
    response_row, response_col = response_data.shape

    # Check data for correct shape and type
    if response_row != r:
        msg = "Feature data and response data should have the same number of rows. \n"
        msg += f"Feature data has {r} and response data has {response_row}."
        raise ValueError(msg)
    if response_col != 1:
        msg = f"Response data should have 1 column. Was given {response_col}."
        raise ValueError(msg)
    if not isinstance(feature_data[0, 0], float):
        msg = f"feature_data is {type(feature_data[0, 0])} type. "
        msg += "It should be float type"
        raise TypeError(msg)
    if not isinstance(response_data[0, 0], float):
        msg = f"response_data is {type(response_data[0, 0])} type. "
        msg += "It should be float type"
        raise TypeError(msg)

    if include_second_order and r % (2 * k + 2) == 0:
        n = int(r / (2 * k + 2))
    elif not include_second_order and r % (k + 2) == 0:
        n = int(r / (k + 2))
    else:
        raise ValueError("feature_data is not in the right format")

    y_A = response_data[:n]
    y_B = response_data[n: 2 * n]

    y_AB = np.zeros((n, k))
    S_i = np.zeros(k)
    S_Ti = np.zeros(k)
    for i in range(k):
        y_AB[:, i] = np.squeeze(response_data[(i + 2) * n: (i + 3) * n])
        S_i[i] = np.nanmean(y_B * (y_AB[:, i] - y_A)) / np.nanvar(y_A)
        S_Ti[i] = 0.5 * np.nanmean((y_A - y_AB[:, i]) ** 2) / np.nanvar(y_A)

    if include_second_order:
        y_BA = np.zeros((n, k))
        S_ij = np.zeros((k, k))
        for i in range(k):
            y_BA[:, i] = response_data[(i + k + 2) * n: (i + k + 3) * n]
            for j in range(i + 1, k):
                S_ij[i, j] = np.nanmean(
                    y_BA[:, i] * y_AB[:, j] - y_A * y_B, axis=0
                ) / np.nanvar(y_A)
                S_ij[i, j] = S_ij[i, j] - S_i[i] - S_i[j]
        return S_i, S_Ti, S_ij
    else:
        return S_i, S_Ti


def oat_score_function(feature_data, response_data, method='morris', statistic='mu_star', **kwargs):
    """
    Score function that works with both Morris and One-at-a-Time methods.

    Args:
        feature_data: Array of input samples
        response_data: Array of model outputs
        method: 'morris' or 'oat' (one-at-a-time)
        statistic: For Morris: 'mu_star', 'sigma', 'mu'. For OAT: 'mean', 'std', 'max', 'min'
        **kwargs: Additional arguments passed to the underlying functions

    Returns:
        Array of scores for each parameter
    """
    # Filter out 'powers' argument since Morris/OAT don't use it
    filtered_kwargs = {k: v for k, v in kwargs.items() if k != 'powers'}

    if method.lower() == 'morris':
        results = morris_effects(feature_data, response_data, **filtered_kwargs)
        if statistic == 'mu_star':
            return results['mu_star']
        elif statistic == 'sigma':
            return results['sigma']
        elif statistic == 'mu':
            return results['mu']
        else:
            raise ValueError(f"Unknown Morris statistic: {statistic}")

    elif method.lower() in ['oat', 'one_at_a_time']:
        effects = one_at_a_time_effects(feature_data, response_data)  # Shape: (p, 2)
        if statistic == 'mean':
            return np.mean(effects, axis=1)
        elif statistic == 'std':
            return np.std(effects, axis=1, ddof=1)
        elif statistic == 'max':
            return np.max(np.abs(effects), axis=1)  # Max absolute effect
        elif statistic == 'min':
            return np.min(np.abs(effects), axis=1)  # Min absolute effect
        elif statistic == 'range':
            return np.ptp(effects, axis=1)  # Peak-to-peak (range)
        else:
            raise ValueError(f"Unknown OAT statistic: {statistic}")
    else:
        raise ValueError(f"Unknown method: {method}")


def morris_score_plot(ax, feature_data, response_data, feature_names, response_names, 
                      show_both=True, degree=1, interaction_only=True):
    """
    Plots Morris score plot showing mu_star and optionally sigma

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data
        - response_data ([[float]]): Array-like of response data
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - show_both (bool): If True, shows both mu_star and sigma; if False, only mu_star
        - degree (int): Maximum degree of interaction (Morris only works with degree=1)
        - interaction_only (bool): Whether to only include lowest powers of interaction
    """
    # Morris only works with original parameters (degree=1)
    if degree > 1:
        raise ValueError("Morris method only works with degree=1 (no polynomial interactions)")

    # Convert to numpy arrays and handle 1D response data
    feature_data = np.asarray(feature_data)
    response_data = np.asarray(response_data)
    if response_data.ndim == 1:
        response_data = response_data.reshape(-1, 1)

    # Handle single axes case
    if not hasattr(ax, '__iter__'):
        ax = [ax]

    # Calculate Morris statistics
    results = morris_effects(feature_data, response_data.squeeze())
    mu_star = results['mu_star']
    sigma = results['sigma']

    # Create the plot
    for y_col, y_name, axes in zip(response_data.T, response_names, ax):
        if show_both:
            x = np.arange(len(feature_names))
            width = 0.35

            bars1 = axes.bar(x - width/2, mu_star, width, label='μ* (Overall Effect)', alpha=0.8)
            bars2 = axes.bar(x + width/2, sigma, width, label='σ (Interactions)', alpha=0.8)

            axes.set_xlabel('Parameter')
            axes.set_ylabel('Morris Score')
            axes.set_title(f'Morris Sensitivity Analysis ({y_name})')
            axes.set_xticks(x)
            axes.set_xticklabels(feature_names, rotation=70)
            axes.legend()

            # Add value labels on bars
            for bar in bars1:
                height = bar.get_height()
                axes.text(bar.get_x() + bar.get_width()/2., height,
                         f'{height:.3f}', ha='center', va='bottom', fontsize=9)
            for bar in bars2:
                height = bar.get_height()
                axes.text(bar.get_x() + bar.get_width()/2., height,
                         f'{height:.3f}', ha='center', va='bottom', fontsize=9)
        else:
            bars = axes.bar(feature_names, mu_star)
            axes.tick_params(axis='x', labelrotation=70)
            axes.set_xlabel('Parameter')
            axes.set_ylabel('μ* Score')
            axes.set_title(f'Morris μ* (Overall Sensitivity) ({y_name})')

            # Add value labels
            for bar in bars:
                height = bar.get_height()
                axes.text(bar.get_x() + bar.get_width()/2., height,
                         f'{height:.3f}', ha='center', va='bottom', fontsize=9)


def morris_rank_plot(ax, feature_data, response_data, feature_names, response_names,
                     rank_by='mu_star', show_both=False, degree=1, interaction_only=True):
    """
    Plots Morris rank plot

    Args:
        - ax (matplotlib.Axes): Axes to plot to
        - feature_data ([[float]]): Array-like of feature data
        - response_data ([[float]]): Array-like of response data
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - rank_by (string): 'mu_star' or 'sigma' - which statistic to use for ranking
        - show_both (bool): If True, creates separate ranking for both mu_star and sigma
        - degree (int): Maximum degree of interaction (Morris only works with degree=1)
        - interaction_only (bool): Whether to only include lowest powers of interaction
    """
    # Morris only works with original parameters (degree=1)
    if degree > 1:
        raise ValueError("Morris method only works with degree=1 (no polynomial interactions)")

    # Convert to numpy arrays and handle 1D response data
    feature_data = np.asarray(feature_data)
    response_data = np.asarray(response_data)
    if response_data.ndim == 1:
        response_data = response_data.reshape(-1, 1)

    # Calculate Morris statistics for all responses
    all_mu_star = []
    all_sigma = []
    for y_col in response_data.T:
        results = morris_effects(feature_data, y_col)
        all_mu_star.append(results['mu_star'])
        all_sigma.append(results['sigma'])

    mu_star_scores = np.array(all_mu_star)
    sigma_scores = np.array(all_sigma)

    if show_both:
        # Show both statistics - extend feature names and scores
        extended_feature_names = [f"{name} (μ*)" for name in feature_names] + [f"{name} (σ)" for name in feature_names]
        scores = np.concatenate([mu_star_scores, sigma_scores], axis=1)
        num_features = len(extended_feature_names)
    else:
        # Show only the requested statistic
        scores = mu_star_scores if rank_by == 'mu_star' else sigma_scores
        extended_feature_names = feature_names
        num_features = len(feature_names)

    num_responses = len(response_names)

    # Create rank plot manually
    order = np.argsort(-scores)
    rank_table = np.argsort(order)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)

    color_mesh = ax.pcolormesh(num_features - rank_table.T, edgecolors='None', lw=0.0, cmap=cm.plasma,
                               vmin=0, vmax=num_features)

    ax.hlines(np.arange(1, num_features + 1), 0, num_responses, color='k')
    ax.vlines(np.arange(1, num_responses + 1), 0, num_features, color='k')

    cbar = plt.colorbar(color_mesh, cax=cax, orientation='vertical')

    for response_idx in range(num_responses):
        for feature_idx in range(num_features):
            rank_value = rank_table[response_idx, feature_idx] + 1
            ax.text(response_idx + 0.5, feature_idx + 0.5, rank_value, 
                   ha='center', va='center')

    ax.set_xticks(np.arange(num_responses) + 0.5)
    ax.set_xticklabels(response_names, rotation=35, ha='right')
    ax.set_xlabel('Outputs', ha='center')

    ax.set_yticks(np.arange(num_features) + 0.5)
    ax.set_yticklabels(extended_feature_names)
    ax.set_ylabel('Parameters')

    cbar_labels = np.arange(1, num_features + 1).astype('str')
    cbar_labels[-1] += ' (Least Sensitive)'
    cbar_labels[0] += ' (Most Sensitive)'

    cbar.set_ticks(num_features - np.arange(num_features))
    cbar.set_ticklabels(cbar_labels)

    ax.set_xlim([0, num_responses])
    ax.set_ylim([0, num_features])
    ax.invert_yaxis()


def oat_score_plot(ax, feature_data, response_data, feature_names, response_names,
                   statistic='max'):
    """
    Plots One-at-a-Time score plot
    
    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data
        - response_data ([[float]]): Array-like of response data
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - statistic (string): 'mean', 'std', 'max', 'min', or 'range'
    """
    # Convert to numpy arrays and handle 1D response data
    feature_data = np.asarray(feature_data)
    response_data = np.asarray(response_data)
    if response_data.ndim == 1:
        response_data = response_data.reshape(-1, 1)

    # Handle single axes case
    if not hasattr(ax, '__iter__'):
        ax = [ax]

    title = f'OAT {statistic.title()} Effects'
    y_label = f'{statistic} Effect'

    # Create the plot directly
    for y_col, y_name, axes in zip(response_data.T, response_names, ax):
        scores = oat_score_function(feature_data, y_col, method='oat', statistic=statistic)
        
        axes.bar(feature_names, scores)
        axes.tick_params(axis='x', labelrotation=70)
        axes.set_xlabel('Parameter')
        axes.set_ylabel(y_label)
        axes.set_title('{} ({})'.format(title, y_name))


def oat_rank_plot(ax, feature_data, response_data, feature_names, response_names,
                  statistic='max'):
    """
    Plots One-at-a-Time rank plot

    Args:
        - ax (matplotlib.Axes): Axes to plot to
        - feature_data ([[float]]): Array-like of feature data
        - response_data ([[float]]): Array-like of response data
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - statistic (string): 'mean', 'std', 'max', 'min', or 'range'
    """
    # Convert to numpy arrays and handle 1D response data
    feature_data = np.asarray(feature_data)
    response_data = np.asarray(response_data)
    if response_data.ndim == 1:
        response_data = response_data.reshape(-1, 1)

    # Calculate OAT statistics for all responses using the existing function
    all_scores = []
    for y_col in response_data.T:
        scores = oat_score_function(feature_data, y_col, method='oat', statistic=statistic)
        all_scores.append(scores)

    scores = np.array(all_scores)
    num_features = len(feature_names)
    num_responses = len(response_names)

    # Create rank plot manually
    order = np.argsort(-scores)
    rank_table = np.argsort(order)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    color_mesh = ax.pcolormesh(num_features - rank_table.T, edgecolors='None', lw=0.0, cmap=cm.plasma)

    ax.hlines(np.arange(1, num_features + 1), 0, num_responses, color='k')
    ax.vlines(np.arange(1, num_responses + 1), 0, num_features, color='k')

    cbar = plt.colorbar(color_mesh, cax=cax, orientation='vertical')

    for i in range(num_responses * num_features):
        idx = np.unravel_index(i, (num_responses, num_features))
        ax.text(idx[0] + .5, idx[1] + .5, rank_table[idx] + 1, ha='center', va='center')

    ax.set_xticks(np.arange(num_responses) + 0.5)
    ax.set_xticklabels(response_names, rotation=35, ha='right')
    ax.set_xlabel('Outputs', ha='center')

    ax.set_yticks(np.arange(num_features) + 0.5)
    ax.set_yticklabels(feature_names)
    ax.set_ylabel('Parameters')

    cbar_labels = np.arange(1, num_features + 1).astype('str')
    cbar_labels[-1] += ' (Least Sensitive)'
    cbar_labels[0] += ' (Most Sensitive)'

    cbar.set_ticks(num_features - np.arange(num_features))
    cbar.set_ticklabels(cbar_labels)

    ax.set_xlim([0, num_responses])
    ax.set_ylim([0, num_features])
    ax.invert_yaxis()


def sobol_score_plot(ax, feature_data, response_data, feature_names, response_names,
                     index_type='first_order', include_second_order=False):
    """
    Plots Sobol indices score plot

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data
        - response_data ([[float]]): Array-like of response data
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - index_type (string): 'first_order' or 'total_order'
        - include_second_order (bool): Whether to calculate second-order indices
    """
    # Convert to numpy arrays and handle 1D response data
    feature_data = np.asarray(feature_data)
    response_data = np.asarray(response_data)
    if response_data.ndim == 1:
        response_data = response_data.reshape(-1, 1)

    # Calculate Sobol indices
    if include_second_order:
        S_i, S_Ti, S_ij = sobol_indices(feature_data, response_data, include_second_order=True)
    else:
        S_i, S_Ti = sobol_indices(feature_data, response_data, include_second_order=False)

    if index_type == 'first_order':
        scores = S_i
    elif index_type == 'total_order':
        scores = S_Ti
    else:
        raise ValueError(f"Unknown index_type: {index_type}")

    # Handle single axes case
    if not hasattr(ax, '__iter__'):
        ax = [ax]

    title = f'Sobol {index_type.replace("_", " ").title()} Indices'
    y_label = f'{index_type.replace("_", " ").title()} Index'

    # Create the plot directly
    for y_col, y_name, axes in zip(response_data.T, response_names, ax):
        axes.bar(feature_names, scores)
        axes.tick_params(axis='x', labelrotation=70)
        axes.set_xlabel('Parameter')
        axes.set_ylabel(y_label)
        axes.set_title('{} ({})'.format(title, y_name))


def sobol_rank_plot(ax, feature_data, response_data, feature_names, response_names,
                    index_type='first_order', include_second_order=False):
    """
    Plots Sobol indices rank plot

    Args:
        - ax (matplotlib.Axes): Axes to plot to
        - feature_data ([[float]]): Array-like of feature data
        - response_data ([[float]]): Array-like of response data
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - index_type (string): 'first_order' or 'total_order'
        - include_second_order (bool): Whether to calculate second-order indices
    """
    # Convert to numpy arrays and handle 1D response data
    feature_data = np.asarray(feature_data)
    response_data = np.asarray(response_data)
    if response_data.ndim == 1:
        response_data = response_data.reshape(-1, 1)

    # Calculate Sobol indices for all responses
    all_scores = []
    for y_col in response_data.T:
        y_col_2d = y_col.reshape(-1, 1)  # Sobol expects 2D

        if include_second_order:
            S_i, S_Ti, S_ij = sobol_indices(feature_data, y_col_2d, include_second_order=True)
        else:
            S_i, S_Ti = sobol_indices(feature_data, y_col_2d, include_second_order=False)

        if index_type == 'first_order':
            scores = S_i
        elif index_type == 'total_order':
            scores = S_Ti
        else:
            raise ValueError(f"Unknown index_type: {index_type}")

        all_scores.append(scores)

    scores = np.array(all_scores)
    num_features = len(feature_names)
    num_responses = len(response_names)
    
    # Create rank plot manually
    order = np.argsort(-scores)
    rank_table = np.argsort(order)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    color_mesh = ax.pcolormesh(num_features - rank_table.T, edgecolors='None', lw=0.0, cmap=cm.plasma)

    ax.hlines(np.arange(1, num_features + 1), 0, num_responses, color='k')
    ax.vlines(np.arange(1, num_responses + 1), 0, num_features, color='k')

    cbar = plt.colorbar(color_mesh, cax=cax, orientation='vertical')

    for i in range(num_responses * num_features):
        idx = np.unravel_index(i, (num_responses, num_features))
        ax.text(idx[0] + .5, idx[1] + .5, rank_table[idx] + 1, ha='center', va='center')

    ax.set_xticks(np.arange(num_responses) + 0.5)
    ax.set_xticklabels(response_names, rotation=35, ha='right')
    ax.set_xlabel('Outputs', ha='center')

    ax.set_yticks(np.arange(num_features) + 0.5)
    ax.set_yticklabels(feature_names)
    ax.set_ylabel('Parameters')

    cbar_labels = np.arange(1, num_features + 1).astype('str')
    cbar_labels[-1] += ' (Least Sensitive)'
    cbar_labels[0] += ' (Most Sensitive)'

    cbar.set_ticks(num_features - np.arange(num_features))
    cbar.set_ticklabels(cbar_labels)

    ax.set_xlim([0, num_responses])
    ax.set_ylim([0, num_features])
    ax.invert_yaxis()


def lasso_path_plot(ax, feature_data, response_data, feature_names, response_names,
                    degree=1, method='lasso'):
    """
        Plots Lasso paths

        Plots the path of linear regression coefficients for different amounts of l1 regularization.
        The x-axis is the shrinkage ratio which varies between 0 and 1.
        The shrinkage ratio is a ratio between the L1 norm of the regularized coefficients and the
        L1 norm of the un-regularized coefficients.
        Features that go to zero sooner tend to contribute less to the sensitivity of the response.

        Args:
            - ax ([matplotlib.Axes]): Array-like of Axes to plot to
            - feature_data ([[float]]): Array-like of feature data. Each column is a feature; each
              row is an observation.
            - response_data ([[float]]): Array-like of response data. Rows correspond to rows in
              feature data.
            - feature_names ([string]): Array-like of feature names
            - response_names ([string]): Array-like of response names
            - degree (int): Maximum degree of interaction
            - method (string): Which algorithm to use; lasso: Coordinate descent, lars: least angle
              regression.

        Raises:
            -
    """
    feature_interaction_data, feature_interaction_names, _ = \
        _make_interactions(feature_data, feature_names, degree=degree, interaction_only=False)

    for response_column_data, response_column_name, axes in zip(response_data.T, response_names, ax):

        alphas, active, coefficients = lars_path(feature_interaction_data,
                                                 response_column_data,
                                                 method=method)

        shrinkage = np.sum(np.abs(coefficients.T), axis=1)
        shrinkage /= shrinkage[-1]

        lines = [axes.plot(shrinkage, coefficient, color=np.random.rand(3))[0]
                 for coefficient in coefficients]
        axes.legend(lines, feature_interaction_names)

        for s in shrinkage:
            axes.axvline(s, 0, 1, linestyle='dashed', color='grey')
        axes.hlines(0, 0, 1, linestyles='dashed', color='grey')

        axes.set_xlabel('Shrinkage Factor')
        axes.set_ylabel('Coefficient')
        axes.set_title('{} Path'.format(method.upper()))


def sensitivity_plot(ax, surrogate_model, feature_names, response_names, feature_ranges,
                     num_plot_points=100, num_seed_points=5, seed=2018):
    """
    Plots sensitivity plots

    Also called spaghetti plot.
    Plots a set of 1-d slices of the response surface.
    Colors across plots correspond to the same seed point.

    Args:
        - ax ([[matplotlib.Axes]]): Array-like of Axes to plot to
        - surrogate_model (surrogate_model): Surrogate model which has been fit
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - feature_ranges ([[float]]): Array-like of feature ranges.
                                      Each row is a length 2 array of the lower and upper bounds.
        - num_plot_points (int): Number of points to plot on each dimension sweep
        - num_seed_points (int): Number of points to use as default points
        - seed (int): RNG seed

    Raises:
        -
    """
    rand_gen = np.random.default_rng(seed)

    np_seed_points = LatinHyperCubeSampler.sample_points(num_points=num_seed_points,
                                                         box=feature_ranges,
                                                         seed=seed)
    colors = rand_gen.random((num_seed_points, 3), dtype='float')

    for color, point in zip(colors, np_seed_points):
        for feature_index, feature_name in enumerate(feature_names):

            dimension_sweep = np.linspace(feature_ranges[feature_index][0],
                                          feature_ranges[feature_index][1],
                                          num_plot_points)
            new_feature_data = np.tile(point, reps=(num_plot_points, 1))
            new_feature_data[:, feature_index] = dimension_sweep
            response_prediction = surrogate_model.predict(new_feature_data)

            for response_index, response_name in enumerate(response_names):
                if len(response_names) == 1:
                    ax[feature_index].set_xlabel(feature_name)
                    ax[feature_index].set_ylabel(response_name)
                    ax[feature_index].plot(dimension_sweep, response_prediction, color=color, alpha=.75)
                else:
                    ax[feature_index, response_index].set_xlabel(feature_name)
                    ax[feature_index, response_index].set_ylabel(response_name)
                    ax[feature_index, response_index].plot(dimension_sweep,
                                                           response_prediction[:, response_index],
                                                           color=color,
                                                           alpha=.75)


def f_score_plot(ax, feature_data, response_data, feature_names, response_names,
                 degree=1, interaction_only=True, use_p_value=False):
    """
    Plots F score plot

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - degree (int): Maximum degree of interaction
        - interaction_only (bool): Whether to only include lowest powers of interaction or
          include higher powers.
        - use_p_value (bool): Whether to use p-values or raw F-score

    Raises:
        -
    """
    _score_plot(ax=ax,
                feature_data=feature_data,
                response_data=response_data,
                feature_names=feature_names,
                response_names=response_names,
                degree=degree,
                interaction_only=interaction_only,
                score_function=_f_p_score if use_p_value else _f_score,
                title='F Score',
                y_axis_label='p-value' if use_p_value else 'Score')

    if use_p_value:
        for axes in ax:
            axes.axhline(.05, 0, 1, linestyle='dashed', color='red', label='alpha=.05')


def mutual_info_score_plot(ax, feature_data, response_data, feature_names, response_names, n_neighbors=3):
    """
    Plots mutual information score plot

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - n_neighbors (int): How many neighboring bins to consider when estimating mutual
          information.

    Raises:
        -
    """
    _score_plot(ax=ax,
                feature_data=feature_data,
                response_data=response_data,
                feature_names=feature_names,
                response_names=response_names,
                degree=1,
                interaction_only=False,
                score_function=_mutual_info_score,
                title='Mutual Information',
                y_axis_label='Shared Information (nats)',
                n_neighbors=n_neighbors)


def pce_score_plot(ax, feature_data, response_data, feature_names, response_names, feature_ranges,
                   degree=1, model_degrees=1):
    """
    Plots PCE score plot

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows in
          feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - feature_ranges ([[float]]): Array-like of feature ranges. Each row is a length 2
          array of the lower and upper bounds.
        - degree (int): Maximum degree of interaction to plot
        - model_degrees (int): Maximum degree of interaction for PCE model

    Raises:
        -
    """
    _score_plot(ax=ax,
                feature_data=feature_data,
                response_data=response_data,
                feature_names=feature_names,
                response_names=response_names,
                degree=degree,
                interaction_only=True,
                score_function=_pce_score,
                title='PCE Variance Decomposition',
                y_axis_label='Variance Contribution',
                ranges=feature_ranges,
                pce_degree=degree,
                model_degrees=model_degrees)


def f_score_rank_plot(ax, feature_data, response_data, feature_names, response_names,
                      degree=1, interaction_only=True, use_p_value=False):
    """
    Plots F score rank plot

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - degree (int): Maximum degree of interaction
        - interaction_only (bool): Whether to only include lowest powers of interaction or
          include higher powers.
        - use_p_value (bool): Whether to use p-values or raw F-score

    Raises:
        -
    """
    _rank_plot(ax=ax,
               feature_data=feature_data,
               response_data=response_data,
               feature_names=feature_names,
               response_names=response_names,
               score_function=(lambda *args: -_f_p_score(*args)) if use_p_value else _f_score,
               degree=degree,
               interaction_only=interaction_only)


def mutual_info_rank_plot(ax, feature_data, response_data, feature_names, response_names, n_neighbors=3):
    """
    Plots mutual information rank plot

    Args:
        - ax (matplotlib.Axes): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - n_neighbors (int): How many neighboring bins to consider when estimating mutual
          information.

    Raises:
        -
    """
    _rank_plot(ax=ax,
               feature_data=feature_data,
               response_data=response_data,
               feature_names=feature_names,
               response_names=response_names,
               score_function=_mutual_info_score,
               degree=1,
               interaction_only=False,
               n_neighbors=n_neighbors)


def pce_rank_plot(ax, feature_data, response_data, feature_names, response_names, feature_ranges,
                  degree=1, model_degrees=1):
    """
    Plots PCE rank plot

    Args:
        - ax (matplotlib.Axes): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - feature_ranges ([[float]]): Array-like of feature ranges. Each row is a length
          2 array of the lower and upper bounds.
        - degree (int): Maximum degree of interaction to plot
        - model_degrees (int): Maximum degree of interaction for PCE model

    Raises:
        -
    """
    _rank_plot(ax=ax,
               feature_data=feature_data,
               response_data=response_data,
               feature_names=feature_names,
               response_names=response_names,
               score_function=_pce_score,
               degree=degree,
               interaction_only=True,
               ranges=feature_ranges,
               pce_degree=degree,
               model_degrees=model_degrees)


def f_score_network_plot(ax, feature_data, response_data, feature_names, response_names,
                         degree=2, max_size=10.0, label_size=10, alpha=.5):
    """
    Plots F score network plot

    Args:
        - ax ([matplotlib.Axes]): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows
          in feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - degree (int): Maximum degree of interaction
        - max_size (float): Maximum size of elements in plot. Measured in points
        - label_size (int): Font size of labels. Measured in points
        - alpha (float): Opacity of elements in plot.

    Raises:
        -
    """
    _variance_network_plot(ax=ax,
                           feature_data=feature_data,
                           response_data=response_data,
                           feature_names=feature_names,
                           response_names=response_names,
                           score_function=_f_score,
                           method_label='F score',
                           degree=degree,
                           max_size=max_size,
                           label_size=label_size,
                           alpha=alpha)


def pce_network_plot(ax, feature_data, response_data, feature_names, response_names, feature_ranges,
                     degree=2, model_degrees=2, max_size=10.0, label_size=10, alpha=.5):
    """
    Plots PCE network plot

    Args:
        - ax (matplotlib.Axes): Array-like of Axes to plot to
        - feature_data ([[float]]): Array-like of feature data. Each column is a feature;
          each row is an observation.
        - response_data ([[float]]): Array-like of response data. Rows correspond to rows in
          feature data.
        - feature_names ([string]): Array-like of feature names
        - response_names ([string]): Array-like of response names
        - feature_ranges ([[float]]): Array-like of feature ranges. Each row is a length 2
          array of the lower and upper bounds.
        - degree (int): Maximum degree of interaction to plot
        - model_degrees (int): Maximum degree of interaction for PCE model
        - max_size (float): Maximum size of elements in plot. Measured in points
        - label_size (int): Font size of labels. Measured in points
        - alpha (float): Opacity of elements in plot.

    Raises:
        -
    """
    _variance_network_plot(ax=ax,
                           feature_data=feature_data,
                           response_data=response_data,
                           feature_names=feature_names,
                           response_names=response_names,
                           score_function=_pce_score,
                           method_label='PCE',
                           degree=degree,
                           max_size=max_size,
                           label_size=label_size,
                           alpha=alpha,
                           ranges=feature_ranges,
                           pce_degree=degree,
                           model_degrees=model_degrees)


if __name__ == "__main__":
    pass
